// -*- mode: C++; -*-

/* LibGM: A Geometrical Library
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef GM_UTILS_H
#define GM_UTILS_H

#include "gm_bool.h"
#include "gm_const.h"


// one-argument utility functions

inline double gmAbs (double f)
{
  return (f >= 0) ? f : -f;
}


inline double gmCeil (double f)
{
  return (f == int(f)) ? f : (f > 0) ? double(int(f) + 1) : double(int(f));
}


inline double gmCube (double f)
{
  return f * f * f;
}


inline double gmDegrees (double f)
{
  return f * GM_RADTODEG;
}


inline double gmFloor (double f)
{
  return (f == int(f)) ? f : (f > 0) ? double(int(f)) : double(int(f) - 1);
}


inline double gmInv (double f)
{
  return 1.0 / f;
}


inline gmBool gmIsZero (double f)
{
  return (gmAbs(f) < GM_EPSILON);
}


inline double gmRadians (double f)
{
  return f * GM_DEGTORAD;
}


inline double gmRound (double f)
{
  return (f >= 0) ? double(int(f + 0.5)) : double(- int(0.5 - f));
}


inline double gmSign (double f)
{
  return (f < 0) ? -1.0 : 1.0;
}


inline double gmSmooth (double f)
{
  return (3.0 - 2.0 * f) * f * f;
}


inline double gmSqr (double f)
{
  return f * f;
}


inline double gmTrunc (double f)
{
  return double(int(f));
}


inline double gmZSign (double f)
{
  return (f > 0) ? 1.0 : (f < 0) ? -1.0 : 0.0;
}


// two-argument utility functions

inline gmBool gmFuzEQ (double f, double g)
{
  return (f <= g) ? (f >= g - GM_EPSILON) : (f <= g + GM_EPSILON);
}


inline gmBool gmFuzGEQ (double f, double g)
{
  return (f >= g - GM_EPSILON);
}


inline gmBool gmFuzLEQ (double f, double g)
{
  return (f <= g + GM_EPSILON);
}


inline double gmMax (double f, double g)
{
  return (f > g) ? f : g;
}


inline double gmMin (double f, double g)
{
  return (f < g) ? f : g;
}


inline void gmSwap (double& f, double& g)
{
  double tmp = f;
  f = g;
  g = tmp;
}


inline void gmSwap (int& i, int& j)
{
  int tmp = i;
  i = j;
  j = tmp;
}


// multi-argument utility functions

inline void gmClamp (double &f, double l, double h)
{
  if (f < l)
    f = l;
  if (f > h)
    f = h;
}


inline double gmLerp (double f, double l, double h)
{
  return l + ((h - l) * f );
}


inline double gmMax (double f, double g, double h)
{
  return (f > g) ? gmMax(f, h) : gmMax(g, h);
}


inline double gmMin (double f, double g, double h)
{
  return (f < g) ? gmMin(f, h) : gmMin(g, h);
}


inline double gmSlide (double f, double l, double h)
{
  return (f < 0) ? l : (f > 1) ? h : gmLerp(gmSmooth(f), l, h);
}


#endif // GM_UTILS_H
