#include <list.h>

#include <GL/glut.h>

#include "gm.h"
#include "mtk.h"
#include "mtk_manip.h"

#define A 0.525731112119133606
#define B 0.850650808352039932

int faces[20][3] = {
  {4, 0, 1}, {9, 0, 4}, {5, 9, 4}, {5, 4, 8}, {8, 4, 1},
  {10, 8, 1}, {3, 8, 10}, {3, 5, 8}, {2, 5, 3}, {7, 2, 3},
  {10, 7, 3}, {6, 7, 10}, {11, 7, 6}, {0, 11, 6}, {1, 0, 6},
  {1, 6, 10}, {0, 9, 11}, {11, 9, 2}, {2, 9, 5}, {2, 7, 11}
};

gmPoint3 vtx[12];
mtkCoreGL *core;
mtkCameraId cid;
mtkDefaultTrackBoxManipulator *manip;


void reshape (int w, int h)
{
  core->cameraResize(cid, 0, 0, w, h);
}


void display (void)
{
  core->cameraDraw(cid);
  glutSwapBuffers();
}


void mouse (int button, int state, int x, int y)
{
  // unused parameter
  button=button;

  bool need_redraw=false;
  if (state == GLUT_DOWN)
    need_redraw=core->cameraPointer(cid, x, y, MTK_PICK);
  else
    need_redraw=core->cameraPointer(cid, x, y, MTK_RELEASE);
  if (need_redraw)
    display();
}


void motion (int x, int y)
{
  if (core->cameraPointer(cid, x, y, MTK_MOVE))
    display();
}


void keyboard (unsigned char key, int x, int y)
{
  // unused parameters
  x=y;

  switch (key) {
  case 27:
    exit(0);
  default:
    break;
  }
}


void update ()
{
  core->createElement(1);
  for (int i=0; i<20; i++) {
    core->begin(MTK_POLYGON);
    core->setVertex(vtx[faces[i][0]]);
    core->setVertex(vtx[faces[i][1]]);
    core->setVertex(vtx[faces[i][2]]);
    core->end();
  }
  core->finishElement();
}


void rotate (mtkManipulatorId id, list<mtkId> id_list,
             gmPoint3 p, gmVector3 v, double a, void *data)
{
  // unused parameters
  id=id;
  id_list=id_list;
  data=data;

  gmMatrix4 m;
  m=m.translation(p[0],p[1],p[2]);
  m*=m.rotation(a, v);
  m*=m.translation(-p[0],-p[1],-p[2]);
  for (int i=0; i<12; i++)
    vtx[i]=m.transform(vtx[i]);
  update();
}

void translate (mtkManipulatorId id, list<mtkId> id_list,
                gmVector3 v, void *data)
{
  // unused parameters
  id=id;
  id_list=id_list;
  data=data;

  gmMatrix4 m;
  m=m.translation(v[0],v[1],v[2]);
  for (int i=0; i<12; i++)
    vtx[i]=m.transform(vtx[i]);
  update();
}

void scale (mtkManipulatorId id, list<mtkId> id_list, gmPoint3 p,
                gmVector3 v, double f, void *data)
{
  // unused parameters
  id=id;
  id_list=id_list;
  data=data;

  gmMatrix4 m;
  m=m.translation(p[0],p[1],p[2]);
  m*=m.scaling(f,v);
  m*=m.translation(-p[0],-p[1],-p[2]);

  for (int i=0; i<12; i++)
    vtx[i]=m.transform(vtx[i]);
  update();
}


void init (void)
{
  core=new mtkCoreGL();
  cid=core->createCamera();
  core->cameraInitialize(cid);

  manip=new mtkDefaultTrackBoxManipulator(core);
  mtkDraggerId id=manip->getDraggerId();
  //core->setDraggerSize(id, 0.1);

  vtx[0].assign(-A, 0.0, B);
  vtx[1].assign(A, 0.0, B);
  vtx[2].assign(-A, 0.0, -B);
  vtx[3].assign(A, 0.0, -B);
  vtx[4].assign(0.0, B, A);
  vtx[5].assign(0.0, B, -A);
  vtx[6].assign(0.0, -B, A);
  vtx[7].assign(0.0, -B, -A);
  vtx[8].assign(B, A, 0.0);
  vtx[9].assign(-B, A, 0.0);
  vtx[10].assign(B, -A, 0.0);
  vtx[11].assign(-B, -A, 0.0);

  update();
  manip->registerId(1);
  manip->setRotateCallback(&rotate, 0);
  manip->setTranslateCallback(&translate, 0);
  manip->setScaleCallback(&scale, 0);

  core->setLightType(MTK_LIGHT0, MTK_DIRECTIONAL_LIGHT);
  core->setLightPosition(MTK_LIGHT0, gmPoint3(1.2, 1.0, 0.8));
  core->setLightComponents(MTK_LIGHT0, MTK_DIFFUSE,
                           gmPoint4(1.0, 0.0, 0.0, 1.0));
  core->setLight(MTK_LIGHT0, MTK_ACTIVE);
  core->setLighting(MTK_ACTIVE);
}


int main (int argc, char *argv[])
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
  glutCreateWindow("spinning icosahedron");

  glutReshapeFunc(reshape);
  glutDisplayFunc(display);
  glutMouseFunc(mouse);
  glutMotionFunc(motion);
  glutKeyboardFunc(keyboard);

  init();
  glutMainLoop();
  return 0;
}
