#include <stdio.h>
#include <time.h>

#include "gm.h"
#include "mtk.h"
#include "mtk_manip.h"
#include "../mtkx.h"

mtkCoreGL *core;
mtkCameraId cid;

gmPoint3 cubecenter, wirecenter, oldcenter;
gmPoint3 grabpoint;
GLdouble ang;
int grabbing = 0;

time_t crono;

mtkxGlutWindow *janela;
mtkxWindow::mtkxCursorType mode;
mtkDraggerId manipid;

double rnd(int x)
{
   return ((double)x*(double)rand()/(RAND_MAX+1.0));
}

void update ();
int hit();

void mouse (int button, int state, int x, int y)
{
  if(state == mtkxWindow::MTKX_DOWN)
     core->cameraPointer(cid,x,y,MTK_PICK);
  else
  {
     core->cameraPointer(cid,x,y,MTK_RELEASE);
     if(hit())
        printf("Tempo (Bounding Box): %f seg.\n",difftime(time(NULL),crono));
  }
  janela->mtkxPostRedisplay();
}

void mouse3d (int button, int state, GLdouble x, GLdouble y,GLdouble z)
{
   gmMatrix4 m;
   gmVector3 v;
   gmPoint3 p;

   if(button == mtkxWindow::MTKX_LEFT_BUTTON 
      || button == mtkxWindow::MTKX_MIDDLE_BUTTON)
      if(state == mtkxWindow::MTKX_DOWN)
      {
         m = m.scaling(2,2,2);
         v.assign(1,1,1);
         m *= m.rotation(ang, v);
         m *= m.translation(-cubecenter[0], -cubecenter[1], -cubecenter[2]);

         p = p.assign(x,y,z);
         p = m.transform(p);

         if(p[0] >= -1 && p[0] <= 1
            &&p[1] >= -1 && p[1] <= 1
            &&p[2] >= -1 && p[2] <= 1)
         {
            grabbing++;
            grabpoint.assign(x,y,z);
            update();   
            janela->mtkxPostRedisplay();
         }
      }
      else
      {
         if(grabbing > 0)
         {
            grabbing--;
            if(grabbing == 0 && hit())
               printf("Tempo (Cursor 3D): %f seg.\n",difftime(time(NULL),crono));
         }
         update();   
         janela->mtkxPostRedisplay();
      }

}

void motion3d (GLdouble x, GLdouble y, GLdouble z)
{
   gmPoint3 p;

   core->cameraPointer3D(cid,x,y,z,MTK_MOVE);
   if(grabbing > 0)
   {
      p.assign(x,y,z);
      cubecenter += (p-grabpoint);
      grabpoint = p;

      update();   
      janela->mtkxPostRedisplay();
   }
}

void translate(mtkManipulatorId id, list<mtkId> id_list, gmVector3 v, 
               void *data)
{
   id = id;
   id_list = id_list;
   data = data;

   cubecenter += v;

   update();
   janela->mtkxPostRedisplay();
}


void keyboard (unsigned char key, int x, int y)
{
  // unused parameters
  x=y;

  switch (key) {
  case 27:
    exit(0);
  case 32:
     ang += 10;
     update();
     janela->mtkxPostRedisplay();
     break;
  case 'n':
     cubecenter.assign(rnd(3)-1.0, rnd(3)-1.0, rnd(5)-3.0);
     wirecenter.assign(rnd(3)-1.0, rnd(3)-1.0, rnd(5)-3.0);
     ang = rnd(180)-90;
     oldcenter = cubecenter;
     update();
     core->setSelection(1,false);
     janela->mtkxPostRedisplay();
     crono = time(NULL);
     break;
  case 13:
     cubecenter = oldcenter;
     update();
     core->setSelection(1,false);
     janela->mtkxPostRedisplay();
     crono = time(NULL);
     break;
  case 9:
     if(mode == mtkxWindow::MTKX_CURSOR_2D)
        mode = mtkxWindow::MTKX_CURSOR_3D;
     else
        mode = mtkxWindow::MTKX_CURSOR_2D;
     core->setSelection(1,false);
     janela->mtkxSetCursorType(mode);

  default:
     break;
  }
}

int hit()
{
  gmVector3 v;
  v = cubecenter - wirecenter;
  if(v.length() <= 0.05)
     return  1;
  else
     return 0;
}

void update ()
{
  gmMatrix4 m;
  gmPoint3 p;
  gmVector3 v;

  core->setLightType(MTK_LIGHT0, MTK_DIRECTIONAL_LIGHT);
  core->setLightPosition(MTK_LIGHT0, gmPoint3(1.2, 1.0, 0.8));
  core->setLightComponents(MTK_LIGHT0, MTK_DIFFUSE,
                           gmPoint4(1.0, 1.0, 1.0, 1.0));
  core->setLight(MTK_LIGHT0, MTK_ACTIVE);

  core->setLightType(MTK_LIGHT1, MTK_DIRECTIONAL_LIGHT);
  core->setLightPosition(MTK_LIGHT1, gmPoint3(1.2, 1.0, 0.8));
  core->setLightComponents(MTK_LIGHT1, MTK_DIFFUSE,
                           gmPoint4(0.0, 0.0, 1.0, 1.0));


  core->setLight(MTK_LIGHT1, hit() ? MTK_ACTIVE : MTK_INACTIVE);
  core->setLighting(MTK_ACTIVE);


  m = m.translation(cubecenter[0], cubecenter[1], cubecenter[2]);
  v.assign(1,1,1);
  m *= m.rotation(-ang, v);
  m *= m.scaling(0.5,0.5,0.5);

  core->createElement(1);
  core->begin(MTK_POLYGON);
  p.assign(-1,-1,1);
  core->setVertex(m.transform(p));
  p.assign(1,-1,1);
  core->setVertex(m.transform(p));
  p.assign(1,1,1);
  core->setVertex(m.transform(p));
  p.assign(-1,1,1);
  core->setVertex(m.transform(p));
  core->end();

  core->begin(MTK_POLYGON);
  p.assign(1,-1,1);
  core->setVertex(m.transform(p));
  p.assign(1,-1,-1);
  core->setVertex(m.transform(p));
  p.assign(1,1,-1);
  core->setVertex(m.transform(p));
  p.assign(1,1,1);
  core->setVertex(m.transform(p));
  core->end();

  core->begin(MTK_POLYGON);
  p.assign(1,-1,-1);
  core->setVertex(m.transform(p));
  p.assign(1,1,-1);
  core->setVertex(m.transform(p));
  p.assign(-1,1,-1);
  core->setVertex(m.transform(p));
  p.assign(-1,-1,-1);
  core->setVertex(m.transform(p));
  core->end();

  core->begin(MTK_POLYGON);
  p.assign(-1,-1,1);
  core->setVertex(m.transform(p));
  p.assign(-1,-1,-1);
  core->setVertex(m.transform(p));
  p.assign(-1,1,-1);
  core->setVertex(m.transform(p));
  p.assign(-1,1,1);
  core->setVertex(m.transform(p));
  core->end();

  core->begin(MTK_POLYGON);
  p.assign(1,1,1);
  core->setVertex(m.transform(p));
  p.assign(1,1,-1);
  core->setVertex(m.transform(p));
  p.assign(-1,1,-1);
  core->setVertex(m.transform(p));
  p.assign(-1,1,1);
  core->setVertex(m.transform(p));
  core->end();

  core->begin(MTK_POLYGON);
  p.assign(1,-1,1);
  core->setVertex(m.transform(p));
  p.assign(1,-1,-1);
  core->setVertex(m.transform(p));
  p.assign(-1,-1,-1);
  core->setVertex(m.transform(p));
  p.assign(-1,-1,1);
  core->setVertex(m.transform(p));
  core->end();
  core->finishElement();

  m = m.translation(wirecenter[0], wirecenter[1], wirecenter[2]);
  v.assign(1,1,1);
  m *= m.rotation(-ang, v);
  m *= m.scaling(0.5,0.5,0.5);

  core->createElement(2);
  core->begin(MTK_LINE_STRIP);
  p.assign(-1,-1,1);
  core->setVertex(m.transform(p));
  p.assign(1,-1,1);
  core->setVertex(m.transform(p));
  p.assign(1,1,1);
  core->setVertex(m.transform(p));
  p.assign(-1,1,1);
  core->setVertex(m.transform(p));
  p.assign(-1,-1,1);
  core->setVertex(m.transform(p));
  core->end();

  core->begin(MTK_LINE_STRIP);
  p.assign(1,-1,1);
  core->setVertex(m.transform(p));
  p.assign(1,1,1);
  core->setVertex(m.transform(p));
  p.assign(1,1,-1);
  core->setVertex(m.transform(p));
  p.assign(1,-1,-1);
  core->setVertex(m.transform(p));
  p.assign(1,-1,1);
  core->setVertex(m.transform(p));
  core->end();

  core->begin(MTK_LINE_STRIP);
  p.assign(1,-1,-1);
  core->setVertex(m.transform(p));
  p.assign(-1,-1,-1);
  core->setVertex(m.transform(p));
  p.assign(-1,1,-1);
  core->setVertex(m.transform(p));
  p.assign(1,1,-1);
  core->setVertex(m.transform(p));
  p.assign(1,-1,-1);
  core->setVertex(m.transform(p));
  core->end();

  core->begin(MTK_LINE_STRIP);
  p.assign(-1,-1,1);
  core->setVertex(m.transform(p));
  p.assign(-1,-1,-1);
  core->setVertex(m.transform(p));
  p.assign(-1,1,-1);
  core->setVertex(m.transform(p));
  p.assign(-1,1,1);
  core->setVertex(m.transform(p));
  p.assign(-1,-1,1);
  core->setVertex(m.transform(p));
  core->end();

  core->begin(MTK_LINE_STRIP);
  p.assign(1,1,1);
  core->setVertex(m.transform(p));
  p.assign(1,1,-1);
  core->setVertex(m.transform(p));
  p.assign(-1,1,-1);
  core->setVertex(m.transform(p));
  p.assign(-1,1,1);
  core->setVertex(m.transform(p));
  p.assign(1,1,1);
  core->setVertex(m.transform(p));
  core->end();

  core->begin(MTK_LINE_STRIP);
  p.assign(1,-1,1);
  core->setVertex(m.transform(p));
  p.assign(-1,-1,1);
  core->setVertex(m.transform(p));
  p.assign(-1,-1,-1);
  core->setVertex(m.transform(p));
  p.assign(1,-1,-1);
  core->setVertex(m.transform(p));
  p.assign(1,-1,1);
  core->setVertex(m.transform(p));
  core->end();

  core->finishElement();
}




void init (void)
{
  mtkDefaultBoxManipulator *manip;

  core=new mtkCoreGL();
  cid=core->createCamera();
  manip = new mtkDefaultBoxManipulator(core);
  manipid = manip->getDraggerId();
  core->setBoxFaceColor(manipid,0,1,1,0.1);

  update();
  manip->registerId(1);
  manip->setTranslateCallback(&translate, 0);
}


int main (int argc, char *argv[])
{

  mtkxGlutWindow::mtkxGlutInitWindow(&argc,argv);

  init();

  janela = new mtkxGlutWindow(core, cid, "Coloque o cubo no lugar");

  janela->mtkxKeyboardFunc(keyboard);


  janela->mtkxMouse3DFunc(mouse3d);
  janela->mtkxMouseFunc(mouse);
  janela->mtkxMotion3DFunc(motion3d);
  mode = mtkxWindow::MTKX_CURSOR_2D;
  janela->mtkxSetCursorType(mode);

//  core->cameraInitialize(cid);
  srand(time(NULL));
  cubecenter.assign(rnd(3)-1.0, rnd(3)-1.0, rnd(5)-3.0);
  wirecenter.assign(rnd(3)-1.0, rnd(3)-1.0, rnd(5)-3.0);
  //ang = rnd(180)-90;
  oldcenter = cubecenter;
  ang = 0;
  update();

  crono = time(NULL);
  mtkxGlutWindow::mtkxMainLoop();

  return 0;
}

  
