/* Manipulation Toolkit (MTK)
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros and Flavio N. Fernandes
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <time.h>
#include <stdlib.h>

#include "tmodel.h"


TriModel::TriModel (mtkCoreGL *core)
{
  core_=core;
  cnt_=1;
  srand(time(NULL)%RAND_MAX);
}


TriId TriModel::create ()
{
  Tri t;
  double m;

  t.r=(double)rand()/RAND_MAX;
  t.g=(double)rand()/RAND_MAX;
  t.b=(double)rand()/RAND_MAX;
  for (int i=0; i<3; i++) {
    t.v[i][0]=2.0*((double)rand()/RAND_MAX-0.5);
    t.v[i][1]=2.0*((double)rand()/RAND_MAX-0.5);
    t.v[i][2]=2.0*((double)rand()/RAND_MAX-0.5);
    m=t.v[i].originDistance();
    if (m<GM_EPSILON || m>1.0) {
      i--;
      continue;
    }
    t.v[i][0]/=m;
    t.v[i][1]/=m;
    t.v[i][2]/=m;
  }
  pool_.insert(TriMapPair(cnt_, t));
  update_(cnt_);
  return cnt_++;
}


void TriModel::remove (TriId id)
{
  TriMapIt i;
  if ((i=pool_.find(id)) == pool_.end())
    return;
  core_->deleteElement((*i).first);
  pool_.erase(i);
}


void TriModel::move (TriId id, gmVector3 v)
{
  TriMapIt it;
  if ((it=pool_.find(id)) == pool_.end())
    return;
  for (int i=0; i<3; i++)
    (*it).second.v[i]+=v;
  update_(id);
}


void TriModel::scale (TriId id, gmPoint3 p, gmVector3 v, double f)
{
  TriMapIt it;
  if ((it=pool_.find(id)) == pool_.end())
    return;
  gmMatrix4 m;
  m=m.translation(p[0], p[1], p[2]);
  m*=m.scaling(f, v);
  m*=m.translation(-p[0], -p[1], -p[2]);
  for (int i=0; i<3; i++)
    (*it).second.v[i]=m.transform((*it).second.v[i]);
  update_(id);
}


void TriModel::rotate (TriId id, gmPoint3 p, gmVector3 v, double a)
{
  TriMapIt it;
  if ((it=pool_.find(id)) == pool_.end())
    return;
  gmMatrix4 m;
  m=m.translation(p[0], p[1], p[2]);
  m*=m.rotation(a, v);
  m*=m.translation(-p[0], -p[1], -p[2]);
  for (int i=0; i<3; i++)
    (*it).second.v[i]=m.transform((*it).second.v[i]);
  update_(id);
}


void TriModel::shear (TriId id, gmMatrix4 m)
{
  TriMapIt it;
  if ((it=pool_.find(id)) == pool_.end())
    return;
  for (int i=0; i<3; i++)
    (*it).second.v[i]=m.transform((*it).second.v[i]);
  update_(id);
}


void TriModel::update_ (TriId id)
{
  TriMapIt it;
  if ((it=pool_.find(id)) == pool_.end())
    return;
  Tri *pt;
  pt=&((*it).second);
  core_->setMaterialv(MTK_FRONT_AND_BACK, MTK_AMBIENT,
                      gmPoint4(pt->r, pt->g, pt->b, 1.0));
  core_->setMaterialv(MTK_FRONT_AND_BACK, MTK_DIFFUSE,
                      gmPoint4(pt->r, pt->g, pt->b, 1.0));
  core_->setMaterialv(MTK_FRONT_AND_BACK, MTK_SPECULAR,
                      gmPoint4(1.0, 1.0, 1.0, 1.0));
  core_->setMaterialf(MTK_FRONT_AND_BACK, MTK_SHININESS, 5.0);
  core_->setColor(pt->r, pt->g, pt->b);
  core_->createElement(id);
  core_->begin(MTK_POLYGON);
  core_->setVertex(pt->v[0][0], pt->v[0][1], pt->v[0][2]);
  core_->setVertex(pt->v[1][0], pt->v[1][1], pt->v[1][2]);
  core_->setVertex(pt->v[2][0], pt->v[2][1], pt->v[2][2]);
  core_->end();
  core_->finishElement();
}
