// -*- mode: C++; -*-

/* Manipulation Toolkit (MTK)
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros and Flavio N. Fernandes
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef TMODEL_H
#define TMODEL_H

#include <map.h>

#include "gm.h"
#include "mtk.h"


typedef unsigned int TriId;


class Tri {
public:
  gmPoint3 v[3];
  float r, g, b;
};


class TriModel {
public:
  TriModel (mtkCoreGL *core);

  TriId create ();
  void remove (TriId id);
  void move (TriId id, gmVector3 v);
  void scale (TriId id, gmPoint3 p, gmVector3 v, double f);
  void rotate (TriId id, gmPoint3 p, gmVector3 v, double a);
  void shear (TriId id, gmMatrix4 m);

protected:
  typedef map<TriId, Tri, less<TriId> > TriMap;
  typedef pair<const TriId, Tri> TriMapPair;
  typedef TriMap::iterator TriMapIt;

  void update_ (TriId id);

  mtkCoreGL *core_;
  TriId cnt_;
  TriMap pool_;
};


#endif // TMODEL_H
