/* Manipulation Toolkit (MTK)
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros and Flavio N. Fernandes
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <list.h>
#include <iostream.h>

#include <GL/glut.h>

#include "mtk.h"
#include "mtk_manip.h"
#include "tmodel.h"
#include "tri.h"


/*-------- Global Variables --------*/

mtkCoreGL *core;
mtkCameraId cid;
mtkGuideId axis, grid;

mtkDefaultBoxManipulator *box;
mtkDefaultTrackBoxManipulator *ball;
mtkDefaultJackManipulator *jack;

TriModel *model;

bool moving;

#if 0
list<Camera *> cameras;
#endif


//----------------------------------------------------------------//

#if 0
list<Camera *> Camera::all_;
GLXContext Camera::first_ctxt_=0;


Camera::Camera (Widget parent, int w, int h)
{
  id_=core->createCamera();
  core->cameraResize(id_, 0, 0, w, h);
  all_.push_back(this);

  // create a rendering context
  ctxt_=glXCreateContext(display, info,
                         first_ctxt_, // share display lists
                         true); // try direct rendering
  if (!first_ctxt_)
    first_ctxt_=ctxt_;

  // create a OpenGL drawing area
  Arg args[16];
  int n=0;
  XtSetArg(args[n], GLwNrgba, true); n++;
  XtSetArg(args[n], GLwNvisualInfo, info); n++;
  XtSetArg(args[n], GLwNinstallColormap, false); n++; // no new colormap
  XtSetArg(args[n], XmNtraversalOn, true); n++; // allow keyboard traversal
  XtSetArg(args[n], XmNdepth, info->depth); n++;
  XtSetArg(args[n], XmNwidth, w); n++;
  XtSetArg(args[n], XmNheight, h); n++;
  area_=XtCreateWidget("area", glwMDrawingAreaWidgetClass, parent, args, n);
  XtManageChild(area_);
  XtAddCallback(area_, GLwNginitCallback, initCallback_, (XtPointer)this);
  XtAddCallback(area_, GLwNexposeCallback, exposeCallback_, (XtPointer)this);
  XtAddCallback(area_, GLwNresizeCallback, resizeCallback_, (XtPointer)this);
  XtAddEventHandler(area_,
                    ButtonPressMask | Button1MotionMask |
                    ButtonReleaseMask | KeyPressMask,
                    False, eventHandler_, (XtPointer)this);

  // set colormap for drawing area
  n=0;
  XtSetArg(args[n], XmNcolormap, colormap); n++;
  XtSetValues(area_, args, n);
}


Camera::~Camera ()
{
  // core->destroyCamera(id_); *FIXME*
  all_.remove(this);
  if (ctxt_ != first_ctxt_)
    glXDestroyContext(display, ctxt_);
  XtDestroyWidget(area_);
}


void Camera::draw ()
{
  glXMakeCurrent(display, XtWindow(area_), ctxt_);
  core->cameraDraw(id_);
  glXSwapBuffers(display, XtWindow(area_));
}


void Camera::drawAll ()
{
  for (list<Camera *>::iterator i=all_.begin(); i!=all_.end(); i++)
    (*i)->draw();
}


void Camera::initCallback_ (Widget w, XtPointer client, XtPointer call)
{
  // unused parameters
  w=w;
  call=call;

  ((Camera *)client)->init_();
}


void Camera::init_ ()
{
  glXMakeCurrent(display, XtWindow(area_), ctxt_);
  core->cameraInitialize(id_);
}


void Camera::exposeCallback_ (Widget w, XtPointer client, XtPointer call)
{
  // unused parameters
  w=w;
  call=call;

  ((Camera *)client)->draw();
}


void Camera::resizeCallback_ (Widget w, XtPointer client, XtPointer call)
{
  // unused parameters
  w=w;

  GLwDrawingAreaCallbackStruct *data=(GLwDrawingAreaCallbackStruct *)call;
  ((Camera *)client)->resize_(data->width, data->height);
}


void Camera::resize_ (int width, int height)
{
  core->cameraResize(id_, 0, 0, width, height);
}


void Camera::eventHandler_ (Widget w, XtPointer client, XEvent *event,
                            Boolean *cont)
{
  // unused parameters
  cont=cont;

  ((Camera *)client)->event_(w, event);
}


void Camera::event_ (Widget w, XEvent *event)
{
  KeySym ks;
  bool need_redraw=false;
  mtkCameraProjType t;
  mtkState s;

  glXMakeCurrent(display, XtWindow(area_), ctxt_);
  switch (event->type) {
    case ButtonPress:
      // check if button#1 was pressed
      if (event->xbutton.button == Button1) {
	// set input focus to this widget
        XmProcessTraversal(w, XmTRAVERSE_CURRENT);
        if (event->xbutton.state & ControlMask) {
          core->setSelectionMode(MTK_SEL_TOGGLE);
          box->setState(MTK_INACTIVE);
          ball->setState(MTK_INACTIVE);
          jack->setState(MTK_INACTIVE);
        }
        else if (event->xbutton.state & ShiftMask) {
          core->setSelectionMode(MTK_SEL_INCLUSIVE);
          box->setState(MTK_INACTIVE);
          ball->setState(MTK_INACTIVE);
          jack->setState(MTK_INACTIVE);
        }
        else {
          core->setSelectionMode(MTK_SEL_EXCLUSIVE);
          box->setState(MTK_ACTIVE);
          ball->setState(MTK_ACTIVE);
          jack->setState(MTK_ACTIVE);
        }
        need_redraw=core->cameraPointer(id_, event->xbutton.x,
                                        event->xbutton.y, MTK_PICK);
      }
      break;

    case MotionNotify:
      // check for mouse motion with button#1 pressed
      if (event->xmotion.state & Button1Mask) {
        need_redraw=core->cameraPointer(id_, event->xmotion.x,
                                        event->xmotion.y, MTK_MOVE);
      }
      break;

    case ButtonRelease:
      // check if button#1 was released
      if (event->xbutton.button == Button1) {
        need_redraw=core->cameraPointer(id_, event->xbutton.x,
                                        event->xbutton.y, MTK_RELEASE);
      }
      break;

    case KeyPress:
      // find which key was pressed
      ks=XLookupKeysym((XKeyEvent *)event, 0);
      switch (ks) {
        case XK_a:
	  core->getGuideState(axis, &s);
	  if (s == MTK_ACTIVE)
	    core->setGuideState(axis, MTK_INACTIVE);
          else
	    core->setGuideState(axis, MTK_ACTIVE);
          need_redraw=true;
          break;
        case XK_g:
	  core->getGuideState(grid, &s);
	  if (s == MTK_ACTIVE)
	    core->setGuideState(grid, MTK_INACTIVE);
          else
	    core->setGuideState(grid, MTK_ACTIVE);
	  need_redraw=true;
	  break;
        case XK_l:
          core->cameraLook(id_, core->getAllElements());
          need_redraw=true;
          break;
        case XK_p:
          core->getCameraProjection(id_, &t);
          if (t == MTK_PERSPECTIVE)
            core->setCameraProjection(id_, MTK_PARALLEL);
          else
            core->setCameraProjection(id_, MTK_PERSPECTIVE);
          need_redraw=true;
          break;
        case XK_r:
          core->setRenderType(core->getRenderType() == MTK_SHADED?
                              MTK_WIREFRAME:
                              MTK_SHADED);
          need_redraw=true;
          break;
        case XK_s:
          if(core->getLighting()==MTK_INACTIVE)
            core->setLighting(MTK_ACTIVE);
          else
            core->setLighting(MTK_INACTIVE);
          need_redraw=true;
          break;
        case XK_x:
          if (box->getMode() == MTK_DEFAULT) {
            box->setMode(MTK_ALTERNATE);
            core->setBoxFaceColor(box->getDraggerId(), 1.0, 0.5, 0.5, 0.4);
          }
          else {
            box->setMode(MTK_DEFAULT);
            core->setBoxFaceColor(box->getDraggerId(), 0.0, 1.0, 1.0, 0.4);
          }
          need_redraw=true;
          break;
      }
      break;
  }

  // redisplay all cameras if needed
  if (need_redraw)
    drawAll();
}
#endif


//----------------------------------------------------------------//

/*-------- Misc Functions --------*/

void fatal (char *msg)
{
  cerr << "meta: " << msg << '\n';
  exit(1);
}

/*-------- Modeler Functions --------*/

void initialize ()
{
  // initialize MTK core
  core=new mtkCoreGL();
  cid=core->createCamera();
  core->cameraInitialize(cid);

  // create manipulators
  box=new mtkDefaultBoxManipulator(core);
  ball=new mtkDefaultTrackBoxManipulator(core);
  jack=new mtkDefaultJackManipulator(core);

  // create model
  model=new TriModel(core);

  // create random triangles and register them
  TriId id;
  int i;
  for (i=0; i<2; i++) {
    id=model->create();
    box->registerId(id);
  }
  for (i=0; i<2; i++) {
    id=model->create();
    ball->registerId(id);
  }
  for (i=0; i<2; i++) {
    id=model->create();
    jack->registerId(id);
  }

  // register callbacks for model manipulation
  box->setTranslateCallback(&translateModel, 0);
  box->setScaleCallback(&scaleModel, 0);
  box->setRotateCallback(&rotateModel, 0);
  box->setShearCallback(&shearModel, 0);
  ball->setTranslateCallback(&translateModel, 0);
  ball->setScaleCallback(&scaleModel, 0);
  ball->setRotateCallback(&rotateModel, 0);
  jack->setTranslateCallback(&translateModel, 0);
  jack->setScaleCallback(&scaleModel, 0);
  jack->setRotateCallback(&rotateModel, 0);

  // setup lighting environment
  core->setLightType(MTK_LIGHT0, MTK_DIRECTIONAL_LIGHT);
  core->setLightPosition(MTK_LIGHT0, gmPoint3(1.0, 1.0, -1.0));
  core->setLight(MTK_LIGHT0, MTK_ACTIVE);
  core->setLighting(MTK_ACTIVE);
  core->setLightTwoSide(1.0);

  // create guides
  axis=core->createAxis();
  core->setGuideState(axis, MTK_INACTIVE);
  grid=core->createGrid();
  core->setGuideState(grid, MTK_INACTIVE);
}


void translateModel (mtkManipulatorId id, list<mtkId> id_list,
                     gmVector3 v, void *data)
{
  // unused parameters
  id=id;
  data=data;

  for (list<mtkId>::iterator i=id_list.begin(); i!=id_list.end(); i++)
    model->move(*i, v);
}


void scaleModel (mtkManipulatorId id, list<mtkId> id_list,
                 gmPoint3 p, gmVector3 v, double f, void *data)
{
  // unused parameters
  id=id;
  data=data;

  for (list<mtkId>::iterator i=id_list.begin(); i!=id_list.end(); i++)
    model->scale(*i, p, v, f);
}


void rotateModel (mtkManipulatorId id, list<mtkId> id_list,
                  gmPoint3 p, gmVector3 v, double a, void *data)
{
  // unused parameters
  id=id;
  data=data;

  for (list<mtkId>::iterator i=id_list.begin(); i!=id_list.end(); i++)
    model->rotate(*i, p, v, a);
}


void shearModel (mtkManipulatorId id, list<mtkId> id_list,
                 gmMatrix4 m, void *data)
{
  // unused parameters
  id=id;
  data=data;

  for (list<mtkId>::iterator i=id_list.begin(); i!=id_list.end(); i++)
    model->shear(*i, m);
}


void createCallback ()
{
  TriId id=model->create();
  box->registerId(id);
  display();
}


void cutCallback ()
{
  list<mtkId> sel=core->getAllSelected();
  list<mtkId>::iterator i;
  for (i=sel.begin(); i!=sel.end(); i++) {
    if (box->isRegistered(*i))
      box->unregisterId(*i);
    else if (ball->isRegistered(*i))
      ball->unregisterId(*i);
    else
      jack->unregisterId(*i);
    model->remove(*i);
  }
  display();
}


void selectCallback (int action)
{
  list<mtkId> sel=core->getAllElements();
  list<mtkId>::iterator i;
  for (i=sel.begin(); i!=sel.end(); i++)
    switch (action) {
    case 0:
      if (core->isSelected(*i))
        core->setSelection(*i, false);
      break;
    case 1:
      if (!core->isSelected(*i))
        core->setSelection(*i, true);
      break;
    default:
      if (!core->isSelected(*i))
        core->setSelection(*i, true);
      else
        core->setSelection(*i, false);
      break;
    }
  display();
}


#if 0
void newCameraCallback (Widget w, XtPointer client, XtPointer call)
{
  // unused parameters
  w=w;
  client=client;
  call=call;

  // create a top level shell
  Arg args[32];
  int n=0;
  XtSetArg(args[n], XmNdepth, info->depth); n++;
  XtSetArg(args[n], XmNvisual, info->visual); n++;
  XtSetArg(args[n], XmNcolormap, colormap); n++;
  XtSetArg(args[n], XmNdeleteResponse, XmDESTROY); n++;
  Widget shell=XtAppCreateShell("Camera", "ModelerClass",
                                topLevelShellWidgetClass, display, args, n);
  Widget frame=XmCreateFrame(shell, "frame", NULL, 0);

  // create camera object
  Camera *c=new Camera(frame, 320, 320);
  cameras.push_back(c);

  // tell WM to call a close callback for this shell
  Atom atom=XmInternAtom(display, "WM_DELETE_WINDOW", False);
  XmAddWMProtocolCallback(shell, atom, closeCameraCallback, (XtPointer)c);
  XtManageChild(frame);
  XtRealizeWidget(shell);
}


void closeCameraCallback (Widget w, XtPointer client, XtPointer call)
{
  // unused parameters
  w=w;
  call=call;

  Camera *c=(Camera *)client;
  delete c;
  cout << "bye bye\n";
}
#endif


/*-------- Interface Functions --------*/

void reshape (int width, int height)
{
  core->cameraResize(cid, 0, 0, width, height);
}


void display (void)
{
  core->cameraDraw(cid);
  glutSwapBuffers();
}


void mouse (int button, int state, int x, int y)
{
  if (button != GLUT_LEFT_BUTTON)
    return;

  bool need_redraw=false;
  if (state == GLUT_DOWN) {
    // set selection mode depending on keyboard modifier keys
    int modif=glutGetModifiers();
    if (modif == GLUT_ACTIVE_CTRL) {
      core->setSelectionMode(MTK_SEL_TOGGLE);
      box->setState(MTK_INACTIVE);
      ball->setState(MTK_INACTIVE);
      jack->setState(MTK_INACTIVE);
    }
    else if (modif == GLUT_ACTIVE_SHIFT) {
      core->setSelectionMode(MTK_SEL_INCLUSIVE);
      box->setState(MTK_INACTIVE);
      ball->setState(MTK_INACTIVE);
      jack->setState(MTK_INACTIVE);
    }
    else {
      core->setSelectionMode(MTK_SEL_EXCLUSIVE);
      box->setState(MTK_ACTIVE);
      ball->setState(MTK_ACTIVE);
      jack->setState(MTK_ACTIVE);
    }

    // signal mouse button press
    need_redraw=core->cameraPointer(cid, x, y, MTK_PICK);
    moving=true;
  }
  else {
    // signal mouse button release
    need_redraw=core->cameraPointer(cid, x, y, MTK_RELEASE);
    moving=false;
  }
  if (need_redraw)
    display();
}


void motion (int x, int y)
{
  if (moving) {
    // signal mouse motion
    if (core->cameraPointer(cid, x, y, MTK_MOVE))
      display();
  }
}


void keyboard (unsigned char key, int x, int y)
{
  // unused parameters
  x=y;

  bool need_redraw=false;
  mtkCameraProjType t;
  mtkState s;

  switch (key) {
  case 27:
    exit(0);
  case ' ':
    if (box->getMode() == MTK_DEFAULT) {
      box->setMode(MTK_ALTERNATE);
      core->setBoxFaceColor(box->getDraggerId(), 1.0, 0.5, 0.5, 0.4);
    }
    else {
      box->setMode(MTK_DEFAULT);
      core->setBoxFaceColor(box->getDraggerId(), 0.0, 1.0, 1.0, 0.4);
    }
    need_redraw=true;
    break;
  case 'a':
    core->getGuideState(axis, &s);
    if (s == MTK_ACTIVE)
      core->setGuideState(axis, MTK_INACTIVE);
    else
      core->setGuideState(axis, MTK_ACTIVE);
    need_redraw=true;
    break;
  case 'g':
    core->getGuideState(grid, &s);
    if (s == MTK_ACTIVE)
      core->setGuideState(grid, MTK_INACTIVE);
    else
      core->setGuideState(grid, MTK_ACTIVE);
    need_redraw=true;
    break;
  case 'p':
    core->getCameraProjection(cid, &t);
    if (t == MTK_PERSPECTIVE)
      core->setCameraProjection(cid, MTK_PARALLEL);
    else
      core->setCameraProjection(cid, MTK_PERSPECTIVE);
    need_redraw=true;
    break;
  case 'r':
    core->setRenderType(core->getRenderType() == MTK_SHADED?
                        MTK_WIREFRAME:
                        MTK_SHADED);
    if (core->getLighting() == MTK_INACTIVE)
      core->setLighting(MTK_ACTIVE);
    else
      core->setLighting(MTK_INACTIVE);
    need_redraw=true;
    break;
  default:
    break;
  }
  if (need_redraw)
    display();
}


void menu (int value)
{
  switch (value) {
  case MENU_CREATE:
    createCallback();
    break;
  case MENU_CUT:
    cutCallback();
    break;
  case MENU_SELECT_ALL:
    selectCallback(1);
    break;
  case MENU_QUIT:
    exit(0);
    break;
  default:
    break;
  }
}


/*-------- Main Function --------*/

int main (int argc, char *argv[])
{
  // initialize toolkit
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);

  // create main window
  glutInitWindowSize(320, 320);
  glutCreateWindow("Example Modeler");

  // install toolkit callbacks
  glutReshapeFunc(reshape);
  glutDisplayFunc(display);
  glutMouseFunc(mouse);
  glutMotionFunc(motion);
  glutKeyboardFunc(keyboard);

  // create menu
  glutCreateMenu(menu);
  glutAddMenuEntry("Create Primitive", MENU_CREATE);
  glutAddMenuEntry("Select All", MENU_SELECT_ALL);
  glutAddMenuEntry("Cut Selection", MENU_CUT);
  glutAddMenuEntry("--------", MENU_SEP);
  glutAddMenuEntry("Quit", MENU_QUIT);
  glutAttachMenu(GLUT_RIGHT_BUTTON);

  // initialize MTK
  initialize();

  // go to toolkit main loop
  glutMainLoop();

  // make the compiler happy
  return 0;
}
