// -*- mode: C++; -*-

/* Manipulation Toolkit (MTK)
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros and Flavio N. Fernandes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef MTK_CAMERA_H
#define MTK_CAMERA_H

#include <GL/gl.h>

#include "gm.h"
#include "mtk_general.h"


typedef unsigned int mtkCameraId;

enum mtkCameraMode {MTK_ABSOLUTE, MTK_RELATIVE,  MTK_SPHERICAL};
enum mtkCameraProjType {MTK_PARALLEL, MTK_PERSPECTIVE};


class mtkCamera {
public:
  mtkCamera ();

  mtkCameraId getId ();
  void setState (mtkState state);
  mtkState getState ();

  void setAim (double a, double b, double c, mtkCameraMode mode);
  void getAim (double *a, double *b, double *c, mtkCameraMode mode);
  void setPosition (double a, double b, double c, mtkCameraMode mode);
  void getPosition (double *a, double *b, double *c, mtkCameraMode mode);

  void setNear (double n);
  double getNear ();
  void setFar (double f);
  double getFar ();
  void setWindow (double w);
  double getWindow ();
  void setFOV (double a);
  double getFOV ();
  void setRoll (double a);
  double getRoll ();

  void setProjection (mtkCameraProjType type);
  mtkCameraProjType getProjection ();

  void dolly (double forward);
  void tracking (double up, double right);
  void orbit (double up, double right);
  void pan (double right);
  void tilt (double up);
  void roll (double ccw);
  void zoom (double out);
  void look (gmPoint3 center, double radius);
  void level ();
  void reset ();

  void copy (const mtkCamera& c);
  virtual void setup () = 0;

  // Added by Leonardo Mesquita - 15/02/2001
  void setChanged(bool);
  bool getChanged();

protected:
  static mtkCameraId counter_;

  mtkCameraId id_;
  mtkState state_;

  mtkCameraProjType type_;
  gmPoint3 aim_, pos_;
  gmVector3 up_;
  double far_, near_;
  double window_, fov_, roll_;
  bool need_recalc_;

  // Added by Leonardo Mesquita - 15/02/2001
  bool changed_;
};


class mtkCameraGL: public mtkCamera {
public:
  mtkCameraGL ();

  void resize (int x, int y, int width, int height);
  void setup ();

protected:
  GLint x_, y_;
  GLsizei width_, height_;
};


#endif // MTK_CAMERA_H
