/* Manipulation Toolkit (MTK)
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros and Flavio N. Fernandes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <math.h>
#include <GL/glu.h>

#include "mtk_constraint.h"


mtkConstraintGL::mtkConstraintGL ()
{
  active_axis_=0;
  count_=0;
  x_=y_=0;
  phase_=MTK_SETUP_PHASE;
  type_=MTK_SAME_DEPTH;
  offset_=0.0;
  radius_=1.0;
  front_=true;
  mat_update_=true;
}


mtkConstraintType mtkConstraintGL::getType ()
{
  return type_;
}


mtkConstraintPhase mtkConstraintGL::getPhase ()
{
  return phase_;
}


//----------------------------------------------------------------//

void mtkConstraintGL::setSameDepth ()
{
  phase_=MTK_SETUP_PHASE;
  type_=MTK_SAME_DEPTH;
}


void mtkConstraintGL::setSameDepthHorizontal ()
{
  phase_=MTK_SETUP_PHASE;
  type_=MTK_SAME_DEPTH_HORIZONTAL;
}


void mtkConstraintGL::setSameDepthVertical ()
{
  phase_=MTK_SETUP_PHASE;
  type_=MTK_SAME_DEPTH_VERTICAL;
}


void mtkConstraintGL::setSnapAxis (gmPoint3 p, gmVector3 v)
{
  phase_=MTK_SETUP_PHASE;
  type_=MTK_SNAP_AXIS;
  center_=p;
  axis_[0]=v.normalize();
  active_axis_=0;
}


void mtkConstraintGL::setSnapAxisD (gmPoint3 p, gmVector3 v)
{
  phase_=MTK_SETUP_PHASE;
  type_=MTK_SNAP_AXIS_D;
  center_=p;
  axis_[0]=v;
  length_[0]=v.length();
  active_axis_=0;
}


void mtkConstraintGL::setSnapParallelAxis (gmVector3 v)
{
  phase_=MTK_SETUP_PHASE;
  type_=MTK_SNAP_PARALLEL_AXIS;
  axis_[0]=v.normalize();
  active_axis_=0;
}


void mtkConstraintGL::setSnapParallelAxisD (gmVector3 v)
{
  phase_=MTK_SETUP_PHASE;
  type_=MTK_SNAP_PARALLEL_AXIS_D;
  axis_[0]=v;
  length_[0]=v.length();
  active_axis_=0;
}


void mtkConstraintGL::setSnapNearestAxis (gmPoint3 p, gmVector3 v1,
                                         gmVector3 v2, gmVector3 v3)
{
  phase_=MTK_SETUP_PHASE;
  type_=MTK_SNAP_NEAREST_AXIS;
  center_=p;
  axis_[0]=v1.normalize();
  axis_[1]=v2.normalize();
  axis_[2]=v3.normalize();
}


void mtkConstraintGL::setSnapNearestAxisD (gmPoint3 p, gmVector3 v1,
                                          gmVector3 v2, gmVector3 v3)
{
  phase_=MTK_SETUP_PHASE;
  type_=MTK_SNAP_NEAREST_AXIS_D;
  center_=p;
  axis_[0]=v1;
  length_[0]=v1.length();
  axis_[1]=v2;
  length_[1]=v2.length();
  axis_[2]=v3;
  length_[2]=v3.length();
}


void mtkConstraintGL::setSnapNearestParallelAxis (gmVector3 v1, gmVector3 v2,
                                                 gmVector3 v3)
{
  phase_=MTK_SETUP_PHASE;
  type_=MTK_SNAP_NEAREST_PARALLEL_AXIS;
  axis_[0]=v1.normalize();
  axis_[1]=v2.normalize();
  axis_[2]=v3.normalize();
}


void mtkConstraintGL::setSnapNearestParallelAxisD (gmVector3 v1, gmVector3 v2,
                                                  gmVector3 v3)
{
  phase_=MTK_SETUP_PHASE;
  type_=MTK_SNAP_NEAREST_PARALLEL_AXIS_D;
  axis_[0]=v1;
  length_[0]=v1.length();
  axis_[1]=v2;
  length_[1]=v2.length();
  axis_[2]=v3;
  length_[2]=v3.length();
}


void mtkConstraintGL::setSnapPlane (gmVector3 n, double d)
{
  phase_=MTK_SETUP_PHASE;
  type_=MTK_SNAP_PLANE;
  normal_=n.normalize();
  offset_=d;
}


void mtkConstraintGL::setSnapPlane (gmPoint3 p, gmVector3 v1, gmVector3 v2)
{
  phase_=MTK_SETUP_PHASE;
  type_=MTK_SNAP_PLANE;
  normal_=(gmCross(v1, v2)).normalize();
  offset_=-(normal_[0]*p[0]+normal_[1]*p[1]+normal_[2]*p[2]);
}


void mtkConstraintGL::setSnapPlaneD (gmPoint3 p, gmVector3 v1, gmVector3 v2)
{
  phase_=MTK_SETUP_PHASE;
  type_=MTK_SNAP_PLANE_D;
  center_=p;
  axis_[0]=v1;
  length_[0]=v1.length();
  axis_[1]=v2;
  length_[1]=v2.length();
  normal_=(gmCross(v1, v2)).normalize();
  offset_=-(normal_[0]*p[0]+normal_[1]*p[1]+normal_[2]*p[2]);
}


void mtkConstraintGL::setSnapParallelPlane (gmVector3 n)
{
  phase_=MTK_SETUP_PHASE;
  type_=MTK_SNAP_PARALLEL_PLANE;
  normal_=n.normalize();
}


void mtkConstraintGL::setSnapParallelPlane (gmVector3 v1, gmVector3 v2)
{
  phase_=MTK_SETUP_PHASE;
  type_=MTK_SNAP_PARALLEL_PLANE;
  normal_=(gmCross(v1, v2)).normalize();
}


void mtkConstraintGL::setSnapParallelPlaneD (gmVector3 v1, gmVector3 v2)
{
  phase_=MTK_SETUP_PHASE;
  type_=MTK_SNAP_PARALLEL_PLANE_D;
  axis_[0]=v1;
  length_[0]=v1.length();
  axis_[1]=v2;
  length_[1]=v2.length();
  normal_=(gmCross(v1, v2)).normalize();
}


void mtkConstraintGL::setSnapSphere (gmPoint3 c, double r)
{
  phase_=MTK_SETUP_PHASE;
  type_=MTK_SNAP_SPHERE;
  center_=c;
  radius_=r;
}


//----------------------------------------------------------------//

void mtkConstraintGL::edgeIntersect (int x, int y, gmPoint3 p1, gmPoint3 p2,
                                     gmPoint3 *i)
{
  gmPoint3 l1, l2;

  loadMatrices_();
  y=viewport_[3]-y; // flip y
  l1.assign(x, y, 0.0);
  l2.assign(x, y, 1.0);
  unproject_(l1, &l1);
  unproject_(l2, &l2);

  // find nearest point in edge regarding a line perpendicular to window
  if (!gmNearestPoint(p1, p2-p1, l1, l2-l1, i)) {
    // edge and line are parallel; pick the nearest vertex
    project_(p1, &p1);
    project_(p2, &p2);
    *i=(p1[2] < p2[2])? p1: p2;
  }
}


void mtkConstraintGL::faceIntersect (int x, int y,
                                     gmPoint3 p1, gmPoint3 p2, gmPoint3 p3,
                                     gmPoint3 *i)
{
  gmPoint3 p, w1, w2, w3, w4;
  gmVector3 n, v;
  double d;

  loadMatrices_();
  y=viewport_[3]-y; // flip y
  project_(p1, &w1);
  project_(p2, &w2);
  project_(p3, &w3);

  // intersect projected face against a line perpendicular to window
  gmCalcPlane(w1, w2, w3, &n, &d);
  p.assign(x, y, 0.0);
  v.assign(0.0, 0.0, 1.0);
  if (!gmIntersectPlane(p, v, n, d, &w4)) {
    // face is perpendicular to window; pick nearest vertex *!*
    if (w2[2] < w3[2])
      w4=(w1[2] < w2[2])? w1: w2;
    else
      w4=(w1[2] < w3[2])? w1: w3;
  }
  unproject_(w4, i);
}


void mtkConstraintGL::sphereIntersect (int x, int y, gmPoint3 c, double r,
                                       gmPoint3 *i)
{
  gmPoint3 p1, p2;

  loadMatrices_();
  y=viewport_[3]-y; // flip y
  p1.assign(x, y, 0.0);
  p2.assign(x, y, 1.0);
  unproject_(p1, &p1);
  unproject_(p2, &p2);

  // intersect given sphere against a line perpendicular to window
  gmIntersectSphere(p1, p2-p1, c, r, true, i);
}


//----------------------------------------------------------------//

void mtkConstraintGL::start (int x, int y, gmPoint3 p)
{
  gmPoint3 wref;
  phase_=MTK_START_PHASE;
  loadMatrices_();
  count_=0;
  x_=x;
  y_=viewport_[3]-y; // flip y
  object_=p;
  project_(object_, &window_);
  switch (type_) {
    case MTK_SAME_DEPTH:
    case MTK_SAME_DEPTH_HORIZONTAL:
    case MTK_SAME_DEPTH_VERTICAL:
      break;

    case MTK_SNAP_AXIS:
    case MTK_SNAP_AXIS_D:
      break;

    case MTK_SNAP_PARALLEL_AXIS:
    case MTK_SNAP_PARALLEL_AXIS_D:
      center_=object_;
      break;

    case MTK_SNAP_NEAREST_AXIS:
    case MTK_SNAP_NEAREST_AXIS_D:
      break;

    case MTK_SNAP_NEAREST_PARALLEL_AXIS:
    case MTK_SNAP_NEAREST_PARALLEL_AXIS_D:
      center_=object_;
      break;

    case MTK_SNAP_PLANE:
    case MTK_SNAP_PLANE_D:
      break;

    case MTK_SNAP_PARALLEL_PLANE_D:
      center_=object_;
      // fall through...

    case MTK_SNAP_PARALLEL_PLANE:
      offset_=-(normal_[0]*object_[0]+normal_[1]*object_[1]+
                normal_[2]*object_[2]);
      break;

    case MTK_SNAP_SPHERE:
      project_(center_, &wref);
      if (window_[2] <= wref[2])
        front_=true;
      else
        front_=false;
      break;
  }
}


bool mtkConstraintGL::next (int x, int y, gmPoint3 *p, gmVector3 *v)
{
  phase_=MTK_NEXT_PHASE;

  double max, tmp1, tmp2;
  gmPoint3 p1, p2, p3;
  gmVector3 v1;
  bool ret=true;

  y=viewport_[3]-y; // flip y
  count_++;
  *p=object_;
  switch (type_) {
    case MTK_SAME_DEPTH: // calculation in window coordinates
      // add a constant z displacement to last point
      v1.assign(x-x_, y-y_, 0.0);
      window_+=v1;
      unproject_(window_, &p1);
      *v=p1-object_;
      object_=p1;
      break;

    case MTK_SAME_DEPTH_HORIZONTAL: // calculation in window coordinates
      // add a horizontal displacement to last point
      v1.assign(x-x_, 0.0, 0.0);
      window_+=v1;
      unproject_(window_, &p1);
      *v=p1-object_;
      object_=p1;
      break;

    case MTK_SAME_DEPTH_VERTICAL: // calculation in window coordinates
      // add a vertical displacement to last point
      v1.assign(0.0, y-y_, 0.0);
      window_+=v1;
      unproject_(window_, &p1);
      *v=p1-object_;
      object_=p1;
      break;

    case MTK_SNAP_NEAREST_AXIS:
    case MTK_SNAP_NEAREST_PARALLEL_AXIS: // calculation in window coordinates
      if (count_ == 1) {
        // project each given axis and compare with the 2d displacement
        v1.assign(x-x_, y-y_, 0.0);
        project_(center_, &p1);
        p1[2]=0.0;

        // first axis
        project_(center_+axis_[0], &p2);
        p2[2]=0.0;
        max=gmAbs(gmDot((p2-p1).normalize(), v1));
        active_axis_=0;

        // second axis
        project_(center_+axis_[1], &p2);
        p2[2]=0.0;
        tmp1=gmAbs(gmDot((p2-p1).normalize(), v1));
        if (tmp1 > max) {
          max=tmp1;
          active_axis_=1;
        }

        // third axis
        project_(center_+axis_[2], &p2);
        p2[2]=0.0;
        tmp1=gmAbs(gmDot((p2-p1).normalize(), v1));
        if (tmp1 > max)
          active_axis_=2;
      }
      // fall through...

    case MTK_SNAP_AXIS:
    case MTK_SNAP_PARALLEL_AXIS: // calculation in object coordinates
      // find nearest point in axis regarding a line perpendicular to window
      p1.assign(x, y, 0.0);
      p2.assign(x, y, 1.0);
      unproject_(p1, &p1);
      unproject_(p2, &p2);
      if (!gmNearestPoint(center_, axis_[active_axis_], p1, p2-p1, &p3)) {
        // line and axis are parallel
        ret=false;
        break;
      }
      *v=p3-object_;
      object_=p3;
      project_(p3, &window_);
      break;

    case MTK_SNAP_NEAREST_AXIS_D:
    case MTK_SNAP_NEAREST_PARALLEL_AXIS_D: // calculation in window coordinates
      if (count_ == 1) {
        // project each given axis and compare with the 2d displacement
        v1.assign(x-x_, y-y_, 0.0);
        project_(center_, &p1);

        // first axis
        project_(center_+axis_[0], &p2);
        max=gmAbs(gmDot((p2-p1).normalize(), v1));
        active_axis_=0;

        // second axis
        project_(center_+axis_[1], &p2);
        tmp1=gmAbs(gmDot((p2-p1).normalize(), v1));
        if (tmp1 > max) {
          max=tmp1;
          active_axis_=1;
        }

        // third axis
        project_(center_+axis_[2], &p2);
        tmp1=gmAbs(gmDot((p2-p1).normalize(), v1));
        if (tmp1 > max)
          active_axis_=2;
      }
      // fall through...

    case MTK_SNAP_AXIS_D:
    case MTK_SNAP_PARALLEL_AXIS_D: // calculation in object coordinates
      // find nearest point in axis regarding a line perpendicular to window
      p1.assign(x, y, 0.0);
      p2.assign(x, y, 1.0);
      unproject_(p1, &p1);
      unproject_(p2, &p2);
      if (!gmNearestPoint(center_, axis_[active_axis_], p1, p2-p1, &p3)) {
        // line and axis are parallel
        ret=false;
        break;
      }

      // calculate distance to the center and round it to integer steps
      tmp1=gmRound(gmDistance(p3, center_)/length_[active_axis_]);
      tmp1*=gmSign(gmDot(p3-center_, axis_[active_axis_]));
      p3=axis_[active_axis_]*tmp1+center_;
      *v=p3-object_;
      object_=p3;
      project_(p3, &window_);
      break;

    case MTK_SNAP_PARALLEL_PLANE:
    case MTK_SNAP_PLANE: // calculation in object coordinates
      // intersect plane against a line perpendicular to window
      p1.assign(x, y, 0.0);
      p2.assign(x, y, 1.0);
      unproject_(p1, &p1);
      unproject_(p2, &p2);
      if (!gmIntersectPlane(p1, p2-p1, normal_, offset_, &p3)) {
        // plane is perpendicular to window
        ret=false;
        break;
      }
      *v=p3-object_;
      object_=p3;
      project_(p3, &window_);
      break;

    case MTK_SNAP_PARALLEL_PLANE_D:
    case MTK_SNAP_PLANE_D: // calculation in object coordinates
      // intersect plane against a line perpendicular to window
      p1.assign(x, y, 0.0);
      p2.assign(x, y, 1.0);
      unproject_(p1, &p1);
      unproject_(p2, &p2);
      if (!gmIntersectPlane(p1, p2-p1, normal_, offset_, &p3)) {
        // plane is perpendicular to window
        ret=false;
        break;
      }

      // calculate point proj. in both axis and round them to integer steps
      v1=p3-center_;
      tmp1=gmRound(gmDot(v1, axis_[0])/gmSqr(length_[0]));
      tmp2=gmRound(gmDot(v1, axis_[1])/gmSqr(length_[1]));
      p3=axis_[0]*tmp1+axis_[1]*tmp2+center_;
      *v=p3-object_;
      object_=p3;
      project_(p3, &window_);
      break;

    case MTK_SNAP_SPHERE: // calculation in object coordinates
      // intersect sphere against a line perpendicular to window
      p1.assign(x, y, 0.0);
      p2.assign(x, y, 1.0);
      unproject_(p1, &p1);
      unproject_(p2, &p2);
      gmIntersectSphere(p1, p2-p1, center_, radius_, front_, &p3);
      *v=p3-object_;
      object_=p3;
      project_(p3, &window_);
      break;
  }
  x_=x;
  y_=y;
  return ret;
}


void mtkConstraintGL::end (gmPoint3 *p)
{
  phase_=MTK_SETUP_PHASE;
  mat_update_=true;
  *p=object_;
}


void mtkConstraintGL::loadMatrices_ ()
{
  if (mat_update_) {
    glGetDoublev(GL_MODELVIEW_MATRIX, model_);
    glGetDoublev(GL_PROJECTION_MATRIX, proj_);
    glGetIntegerv(GL_VIEWPORT, viewport_);
    mat_update_=false;
  }
}


void mtkConstraintGL::project_ (gmPoint3 op, gmPoint3 *wp)
{
  GLdouble xx, yy, zz;
  gluProject(op[0], op[1], op[2], model_, proj_, viewport_, &xx, &yy, &zz);
  (*wp).assign(xx, yy, zz);
}


void mtkConstraintGL::unproject_ (gmPoint3 wp, gmPoint3 *op)
{
  GLdouble xx, yy, zz;
  gluUnProject(wp[0], wp[1], wp[2], model_, proj_, viewport_, &xx, &yy, &zz);
  (*op).assign(xx, yy, zz);
}
