// -*- mode: C++; -*-

/* Manipulation Toolkit (MTK)
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros and Flavio N. Fernandes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef MTK_CONSTRAINT_H
#define MTK_CONSTRAINT_H

#include <GL/gl.h>

#include "gm.h"
#include "mtk_general.h"


enum mtkConstraintPhase {MTK_SETUP_PHASE, MTK_START_PHASE, MTK_NEXT_PHASE};

enum mtkConstraintType {MTK_SAME_DEPTH,
                       MTK_SAME_DEPTH_HORIZONTAL, MTK_SAME_DEPTH_VERTICAL,
                       MTK_SNAP_AXIS,
                       MTK_SNAP_AXIS_D,
                       MTK_SNAP_PARALLEL_AXIS,
                       MTK_SNAP_PARALLEL_AXIS_D,
                       MTK_SNAP_NEAREST_AXIS,
                       MTK_SNAP_NEAREST_AXIS_D,
                       MTK_SNAP_NEAREST_PARALLEL_AXIS,
                       MTK_SNAP_NEAREST_PARALLEL_AXIS_D,
                       MTK_SNAP_PLANE,
                       MTK_SNAP_PLANE_D,
                       MTK_SNAP_PARALLEL_PLANE,
                       MTK_SNAP_PARALLEL_PLANE_D,
                       MTK_SNAP_SPHERE};

class mtkConstraintGL {
public:
  mtkConstraintGL ();

  mtkConstraintPhase getPhase ();
  mtkConstraintType getType ();

  void setSameDepth ();
  void setSameDepthHorizontal ();
  void setSameDepthVertical ();

  void setSnapAxis (gmPoint3 p, gmVector3 v);
  void setSnapAxisD (gmPoint3 p, gmVector3 v);
  void setSnapParallelAxis (gmVector3 v);
  void setSnapParallelAxisD (gmVector3 v);
  void setSnapNearestAxis (gmPoint3 p, gmVector3 v1, gmVector3 v2,
                           gmVector3 v3);
  void setSnapNearestAxisD (gmPoint3 p, gmVector3 v1, gmVector3 v2,
                            gmVector3 v3);
  void setSnapNearestParallelAxis (gmVector3 v1, gmVector3 v2, gmVector3 v3);
  void setSnapNearestParallelAxisD (gmVector3 v1, gmVector3 v2, gmVector3 v3);

  void setSnapPlane (gmVector3 n, double d);
  void setSnapPlane (gmPoint3 p, gmVector3 v1, gmVector3 v2);
  void setSnapPlaneD (gmPoint3 p, gmVector3 v1, gmVector3 v2);
  void setSnapParallelPlane (gmVector3 n);
  void setSnapParallelPlane (gmVector3 v1, gmVector3 v2);
  void setSnapParallelPlaneD (gmVector3 v1, gmVector3 v2);

  void setSnapSphere (gmPoint3 c, double r);

  void edgeIntersect (int x, int y, gmPoint3 p1, gmPoint3 p2, gmPoint3 *i);
  void faceIntersect (int x, int y, gmPoint3 p1, gmPoint3 p2, gmPoint3 p3,
                      gmPoint3 *i);
  void sphereIntersect (int x, int y, gmPoint3 c, double r, gmPoint3 *i);

  void start (int x, int y, gmPoint3 p);
  bool next (int x, int y, gmPoint3 *p, gmVector3 *v);
  void end (gmPoint3 *p);

protected:
  void loadMatrices_ ();
  void project_ (gmPoint3 op, gmPoint3 *wp);
  void unproject_ (gmPoint3 wp, gmPoint3 *op);

  mtkConstraintPhase phase_;
  mtkConstraintType type_;
  int active_axis_;
  int count_;
  int x_, y_;
  gmPoint3 object_, window_;

  gmPoint3 center_;
  gmVector3 axis_[3];
  double length_[3];
  gmVector3 normal_;
  double offset_, radius_;
  bool front_;

  bool mat_update_;
  GLdouble model_[16];
  GLdouble proj_[16];
  GLint viewport_[4];
};


#endif // MTK_CONSTRAINT_H
