/* Manipulation Toolkit (MTK)
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros and Flavio N. Fernandes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <GL/gl.h>
#include <GL/glu.h>
#include "mtk_cursor.h"


/*********************************************************************

 mtkCursor

**********************************************************************/

mtkCursor::mtkCursor()
{
   _position = gmPoint3(0.0,0.0,1.0);
   _state = MTK_CURSOR_INACTIVE;
}

void mtkCursorGL::draw()
{
  if (_state == MTK_CURSOR_INACTIVE)
    return;

  GLboolean lighting=glIsEnabled(GL_LIGHTING);
  if (lighting)
    glDisable(GL_LIGHTING);

  // draw vertices
  glColor3f(1.0, 0.0, 0.0 );
  glPointSize(5);

  glBegin(GL_POINTS);
  glVertex3f(_position[0], _position[1], _position[2]);
  glEnd();

  // draw edges
  glColor3f(1.0, 1.0, 1.0);
  glLineWidth(2);

  glBegin(GL_LINES);
  glVertex3f(_position[0], _position[1], _position[2]);
  glVertex3f(_position[0]+1.0, _position[1], _position[2]);
  glEnd();

  glBegin(GL_LINES);
  glVertex3f(_position[0], _position[1], _position[2]);
  glVertex3f(_position[0]-1.0, _position[1], _position[2]);
  glEnd();

  glBegin(GL_LINES);
  glVertex3f(_position[0], _position[1], _position[2]);
  glVertex3f(_position[0], _position[1]+1.0, _position[2]);
  glEnd();

  glBegin(GL_LINES);
  glVertex3f(_position[0], _position[1], _position[2]);
  glVertex3f(_position[0], _position[1]-1.0, _position[2]);
  glEnd();

  glBegin(GL_LINES);
  glVertex3f(_position[0], _position[1], _position[2]);
  glVertex3f(_position[0], _position[1], _position[2]+1.0);
  glEnd();

  glBegin(GL_LINES);
  glVertex3f(_position[0], _position[1], _position[2]);
  glVertex3f(_position[0], _position[1], _position[2]-1.0);
  glEnd();

  if (lighting)
    glEnable(GL_LIGHTING);
}

void mtkCursor::move(gmPoint3 position)
{
   _position = position;
}

gmPoint3 mtkCursor::getPosition()
{
   return _position;
}

mtkCursorState mtkCursor::getState()
{
   return _state;
}

void mtkCursor::setState(mtkCursorState state)
{
   _state = state;
}
