// -*- mode: C++; -*-

/* Manipulation Toolkit (MTK)
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros and Flavio N. Fernandes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "mtk_light.h"


  // Radically changed by Leonardo Mesquita - 16/02/2001


mtkLight::mtkLight ()
{
  int i;

  lighting_state_=MTK_INACTIVE;
  for (i=0; i<=7; i++) {
    light_type_[i]=MTK_POINT_LIGHT;
    light_state_[i]=MTK_INACTIVE;
  }

  size_=1.0;
}


void mtkLight::setType (mtkLightId id, mtkLightType type)
{
  light_type_[id]=type;
}


void mtkLight::setSize (float size)
{
  size_=size;
}


float mtkLight::getSize ()
{
  return size_;
}


//----------------------------------------------------------------//

mtkLightGL::mtkLightGL ()
{
  int i;

  // Added by Leonardo Mesquita - 16/02/2001
  // Load the OpenGL defaults
  light_number_[0]=GL_LIGHT0;
  light_number_[1]=GL_LIGHT1;
  light_number_[2]=GL_LIGHT2;
  light_number_[3]=GL_LIGHT3;
  light_number_[4]=GL_LIGHT4;
  light_number_[5]=GL_LIGHT5;
  light_number_[6]=GL_LIGHT6;
  light_number_[7]=GL_LIGHT7;
  for (i=0; i<=7; i++) {
    position_[i][0] = 0;
    position_[i][1] = 0;
    position_[i][2] = 1;
    position_[i][3] = 0;
    ambient_[i][0] = 0;
    ambient_[i][1] = 0;
    ambient_[i][2] = 0;
    ambient_[i][3] = 1;
    diffuse_[i][0] = 0;
    diffuse_[i][1] = 0;
    diffuse_[i][2] = 0;
    diffuse_[i][3] = 1;
    specular_[i][0] = 0;
    specular_[i][1] = 0;
    specular_[i][2] = 0;
    specular_[i][3] = 1;
    spotdir_[i][0] = 0;
    spotdir_[i][1] = 0;
    spotdir_[i][2] = -1;
    exponent_[i] = 0;
    cutoff_[i] = 180;
    att_constant_[i] = 1;
    att_linear_[i] = 0;
    att_quadratic_[i] = 0;

  }
  diffuse_[0][0] = 1;
  diffuse_[0][1] = 1;
  diffuse_[0][2] = 1;
  specular_[0][0] = 1;
  specular_[0][1] = 1;
  specular_[0][2] = 1;

  globallight_[0] = 0.2;
  globallight_[1] = 0.2;
  globallight_[2] = 0.2;
  globallight_[3] = 1;
  twoside_ = 0;
  localviewer_ = 0;

  draw_ = true;
}


void mtkLightGL::setPosition (mtkLightId id, gmPoint3 position)
{
  position_[id][0]=position[0];
  position_[id][1]=position[1];
  position_[id][2]=position[2];

  if (light_type_[id]==MTK_POINT_LIGHT || light_type_[id]==MTK_SPOT_LIGHT)
    position_[id][3]=1.0;
  else
    position_[id][3]=0.0;
}


void mtkLightGL::setComponents (mtkLightId id, mtkMaterialComponent component,
                                gmPoint4 value)
{
  switch (component) {
  case MTK_AMBIENT:
    ambient_[id][0]=value[0];
    ambient_[id][1]=value[1];
    ambient_[id][2]=value[2];
    ambient_[id][3]=value[3];
    break;
  case MTK_DIFFUSE:
    diffuse_[id][0]=value[0];
    diffuse_[id][1]=value[1];
    diffuse_[id][2]=value[2];
    diffuse_[id][3]=value[3];
    break;
  case MTK_SPECULAR:
    specular_[id][0]=value[0];
    specular_[id][1]=value[1];
    specular_[id][2]=value[2];
    specular_[id][3]=value[3];
    break;
  default:
    break;
  }
}


/// If the light is set as a MTK_SPOT light, sets its attributes,
/// otherwise, return false
bool mtkLightGL::setSpotDir (mtkLightId id, mtkSpotLight spot, gmPoint3 value)
{
  if (light_type_[id]==MTK_SPOT_LIGHT && spot==MTK_SPOT_DIRECTION)
  {
     spotdir_[id][0]=value[0];
     spotdir_[id][1]=value[1];
     spotdir_[id][2]=value[2];
  }
  else
    return false;
  return true;
}


/// If the light is set as a MTK_SPOT light, sets its attributes,
/// otherwise, return false
bool mtkLightGL::setSpot (mtkLightId id, mtkSpotLight spot, float value)
{
  if (light_type_[id] == MTK_SPOT_LIGHT) {
    if (spot == MTK_SPOT_EXPONENT)
      exponent_[id] = value;
    else if (spot == MTK_SPOT_CUTOFF)
      cutoff_[id] = value;
  }
  else
    return false;
  return true;
}


/// If the light is set as a MTK_DIRECTIONAL light, DOESN'T make
/// attenuation and return false, otherwise...
bool mtkLightGL::setAttenuation (mtkLightId id, mtkAttenuation att,
                                 float value)
{
  if (light_type_[id] == MTK_DIRECTIONAL_LIGHT)
    return false;

  switch (att) {
  case MTK_CONSTANT:
    att_constant_[id] = value;
    break;
  case MTK_LINEAR:
    att_linear_[id] = value;
    break;
  case MTK_QUADRATIC:
    att_quadratic_[id] = value;
    break;
  }
  return true;
}


void mtkLightGL::setLight (mtkLightId id, mtkState state)
{
  light_state_[id]=state;
}


mtkState mtkLightGL::getLight (mtkLightId id)
{
  return light_state_[id];
}


void mtkLightGL::setLighting (mtkState state)
{
  lighting_state_=state;
}


mtkState mtkLightGL::getLighting ()
{
  return lighting_state_;
}


void mtkLightGL::setGlobalLight (gmPoint4 value)
{
  globallight_[0]=value[0];
  globallight_[1]=value[1];
  globallight_[2]=value[2];
  globallight_[3]=value[3];
}


void mtkLightGL::setTwoSide (float value)
{
  twoside_ = value;
}


void mtkLightGL::setLocalViewer (float value)
{
  localviewer_ = value;
}


void mtkLightGL::drawStar_ (float *position)
{
  float x, y, z;

  x=position[0];
  y=position[1];
  z=position[2];

  glColor3f(1.0, 1.0, 0.0);
  glBegin(GL_LINES);
  glVertex3f(x+0.5*size_, y, z);
  glVertex3f(x-0.5*size_, y, z);
  glVertex3f(x, y+0.5*size_, z);
  glVertex3f(x, y-0.5*size_, z);
  glVertex3f(x, y, z+0.5*size_);
  glVertex3f(x, y, z-0.5*size_);
  
  glVertex3f(x-0.35*size_, y-0.35*size_, z-0.35*size_);
  glVertex3f(x+0.35*size_, y+0.35*size_, z+0.35*size_);

  glVertex3f(x+0.35*size_, y+0.35*size_, z-0.35*size_);
  glVertex3f(x-0.35*size_, y-0.35*size_, z+0.35*size_);

  glVertex3f(x-0.35*size_, y+0.35*size_, z-0.35*size_);
  glVertex3f(x+0.35*size_, y-0.35*size_, z+0.35*size_);

  glVertex3f(x-0.35*size_, y+0.35*size_, z+0.35*size_);
  glVertex3f(x+0.35*size_, y-0.35*size_, z-0.35*size_);
  
  glEnd();
}


void mtkLightGL::draw ()
{
  int i;
  float position[4];

  if (draw_ && lighting_state_==MTK_ACTIVE) {
    glDisable(GL_LIGHTING);
    for (i=0; i<=7; i++)
      if (light_state_[i]==MTK_ACTIVE &&
          light_type_[i]!=MTK_DIRECTIONAL_LIGHT) {
        glGetLightfv(light_number_[i], GL_POSITION, position);
        drawStar_(position);
      }
    glEnable(GL_LIGHTING);
  }

  // Added by Leonardo Mesquita - 16/02/2001
  if(lighting_state_ == MTK_ACTIVE)
  {
     glLightModelfv(GL_LIGHT_MODEL_AMBIENT, globallight_);
     glLightModelfv(GL_LIGHT_MODEL_TWO_SIDE, &twoside_);
     glLightModelfv(GL_LIGHT_MODEL_LOCAL_VIEWER, &localviewer_);

     for (i=0; i<=7; i++)
        if(light_state_[i] == MTK_ACTIVE)
        {
           glLightfv(light_number_[i], GL_POSITION, position_[i]);
           glLightfv(light_number_[i], GL_AMBIENT, ambient_[i]);
           glLightfv(light_number_[i], GL_DIFFUSE, diffuse_[i]);
           glLightfv(light_number_[i], GL_SPECULAR, specular_[i]);
           glLightfv(light_number_[i], GL_SPOT_DIRECTION, spotdir_[i]);
           glLightfv(light_number_[i], GL_SPOT_EXPONENT, &exponent_[i]);
           glLightfv(light_number_[i], GL_SPOT_CUTOFF, &cutoff_[i]);
           glLightfv(light_number_[i], GL_CONSTANT_ATTENUATION,
                     &att_constant_[i]);
           glLightfv(light_number_[i], GL_LINEAR_ATTENUATION, 
                     &att_linear_[i]);
           glLightfv(light_number_[i], GL_QUADRATIC_ATTENUATION, 
                     &att_quadratic_[i]);
           glEnable(light_number_[i]);
        }
        else
           glDisable(light_number_[i]);
     
     glEnable(GL_LIGHTING);
  }
  else
     glDisable(GL_LIGHTING);
}


void mtkLightGL::setLightDraw (bool draw)
{
  draw_=draw;
}
