// -*- mode: C++; -*-

/* Manipulation Toolkit (MTK)
 * Copyright (C) 1997-1998 Marcelo de G. Malheiros and Flavio N. Fernandes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef MTK_LIGHT_H
#define MTK_LIGHT_H

#include <GL/gl.h>

#include "gm.h"
#include "mtk_general.h"


enum mtkLightId {MTK_LIGHT0, MTK_LIGHT1, MTK_LIGHT2, MTK_LIGHT3,
                 MTK_LIGHT4, MTK_LIGHT5, MTK_LIGHT6, MTK_LIGHT7};

enum mtkLightType {MTK_POINT_LIGHT, MTK_DIRECTIONAL_LIGHT, MTK_SPOT_LIGHT};
enum mtkSpotLight {MTK_SPOT_DIRECTION, MTK_SPOT_EXPONENT, MTK_SPOT_CUTOFF};
enum mtkAttenuation {MTK_CONSTANT, MTK_LINEAR, MTK_QUADRATIC};
enum mtkLightModel {MTK_LIGHT_MODEL_AMBIENT, MTK_LIGHT_MODEL_LOCAL,
                    MTK_LIGHT_MODEL_TWO_SIDE};


class mtkLight {
public:
  mtkLight ();

  void setType (mtkLightId id, mtkLightType type);
  void setSize (float size);
  float getSize ();

  virtual void draw () = 0;

protected:
  mtkState lighting_state_;
  mtkLightType light_type_[8];
  mtkState light_state_[8];
  float size_;
};


class mtkLightGL: public mtkLight {
public:
  mtkLightGL ();

  // local light parameters
  void setLight (mtkLightId id, mtkState state);
  mtkState getLight (mtkLightId id);
  void setPosition (mtkLightId id, gmPoint3 position);
  void setComponents (mtkLightId id, mtkMaterialComponent component,
                      gmPoint4 value);

  bool setSpotDir (mtkLightId id, mtkSpotLight spot, gmPoint3 value);
  bool setSpot (mtkLightId id, mtkSpotLight spot, float value);
  bool setAttenuation (mtkLightId id, mtkAttenuation att, float value);

  // global light parameters
  void setLighting (mtkState state);
  mtkState getLighting ();
  void setGlobalLight (gmPoint4 value);
  void setTwoSide (float value);
  void setLocalViewer (float value);
  void setLightDraw (bool draw);

  void draw ();


private:
  GLenum light_number_[8];
  bool draw_;
  
  // Added by Leonardo Mesquita - 16/02/2001
  float position_[8][4];
  float ambient_[8][4];
  float diffuse_[8][4];
  float specular_[8][4];
  float spotdir_[8][3];
  float exponent_[8];
  float cutoff_[8];
  float att_constant_[8];
  float att_linear_[8];
  float att_quadratic_[8];

  float globallight_[4];
  float twoside_;
  float localviewer_;


  void drawStar_ (float *position);
};


#endif // MTK_LIGHT_H
