// -*- mode: C++; -*-
 
/* Manipulation Toolkit Extension (MTKX)
 * Copyright (C) 2000 Leonardo A. G. Mesquita
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <GL/glut.h>
//#include <X11/Xlib.h>

#include <stdio.h>
#include "mtkx.h"


#define SCALE_FACTOR 0.24

int mtkxWindow::_buttons[3];
mtkxGlutWindow *mtkxGlutWindow::_mtkxGlutWindowList;

/*mtkxGlutLabel *mtkxGlutLabel::_mtkxGlutLabelList;

mtkxLabel::mtkxLabel(char *t, int m)
{
   text = new char[strlen(t)+1];
   strcpy(text,t);
   margin = m;
}

void mtkxLabel::setText(char *t)
{
   delete text;
   text = new char[strlen(t)+1];
   strcpy(text,t);
}

void mtkxLabel::setMargin(int m)
{
   margin = m;
}
*/

mtkxWindow::mtkxWindow (mtkCoreGL *core, mtkCameraId cid)
{
  _CursorType = MTKX_CURSOR_2D;
  _MotionType = MTKX_MOTION_PLANES;
  _core = core;
  _cid = cid;

  _reshapeCallback = NULL;
  _displayCallback = NULL;
  _keyboardCallback = NULL;
  _mouseCallback = NULL;
  _mouse3DCallback = NULL;
  _motionCallback = NULL;
  _motion3DCallback = NULL;
  _passiveMotionCallback = NULL;
  _passiveMotion3DCallback = NULL;
  _motionMapping = _mtkxMappingPlanes;
}

void mtkxWindow::mtkxSetCursorType(mtkxCursorType type)
{
   _CursorType = type;
}

void mtkxWindow::mtkxSetMotionType(mtkxMotionType type)
{
   switch(type)
   {
      case MTKX_MOTION_PLANES:
 	 _motionMapping = _mtkxMappingPlanes;
         break;

      case MTKX_MOTION_IMPLICIT:
 	 _motionMapping = NULL;
         break;

      case MTKX_MOTION_USER:
	 break;
   }
   _MotionType = type;
}

void mtkxWindow::mtkxSetImplicit(mtkxImplicitFunction f)
{
   _implicit = f;
}

void mtkxWindow::mtkxReshapeFunc(mtkxReshapeCallback reshape)
{
   _reshapeCallback = reshape;
}

void mtkxWindow::mtkxDisplayFunc(mtkxDisplayCallback display)
{
   _displayCallback = display;
}

void mtkxWindow::mtkxKeyboardFunc(mtkxKeyboardCallback keyboard)
{
   _keyboardCallback = keyboard;
}

void mtkxWindow::mtkxMouseFunc(mtkxMouseCallback mouse)
{
   _mouseCallback = mouse;
}

void mtkxWindow::mtkxMouse3DFunc(mtkxMouse3DCallback mouse)
{
   _mouse3DCallback = mouse;
}

void mtkxWindow::mtkxMotionFunc(mtkxMotionCallback motion)
{
   _motionCallback = motion;
}


void mtkxWindow::mtkxMotion3DFunc(mtkxMotion3DCallback motion)
{
   _motion3DCallback = motion;
}

void mtkxWindow::mtkxPassiveMotionFunc(mtkxPassiveMotionCallback motion)
{
   _passiveMotionCallback = motion;
}

void mtkxWindow::mtkxPassiveMotion3DFunc(mtkxPassiveMotion3DCallback motion)
{
   _passiveMotion3DCallback = motion;
}

void mtkxWindow::mtkxMappingFunc(mtkxMotionMapping map)
{
   _MotionType = MTKX_MOTION_USER;
   _motionMapping = map;
}

void mtkxWindow::_mtkxMappingImplicit(gmPoint3 p, int x, int y, gmPoint3 &result,
				      mtkxImplicitFunction *implicit)
{
   GLdouble scale;
   GLint view[4];

   glGetIntegerv(GL_VIEWPORT,view);
  
   scale = view[2]*SCALE_FACTOR; 
   result[0] = p[0] + (double)x/scale;
   result[2] = p[2] - (double)y/scale;

   if(implicit != NULL)
      result[1] = (*implicit)(result[0], result[2]);
   else
      result[1] = 0;
}

void mtkxWindow::_mtkxMappingPlanes(gmPoint3 p, int x, int y, gmPoint3 &result)
{
   GLdouble scale;
   GLint view[4];

   GLdouble xx,xy,xz, yx, yy, yz, zx, zy, zz;
   GLdouble model[16], proj[16];
   int right, up, bottom;
   int sright, sup, sbottom;

   glGetDoublev(GL_MODELVIEW_MATRIX,model);
   glGetDoublev(GL_PROJECTION_MATRIX,proj);
   glGetIntegerv(GL_VIEWPORT,view);
   gluProject(1.0, 0.0, 0.0, model, proj, view, &xx, &xy, &xz);
   gluProject(0.0, 1.0, 0.0, model, proj, view, &yx, &yy, &yz);
   gluProject(0.0, 0.0, 1.0, model, proj, view, &zx, &zy, &zz);

   xx -= view[2]/2;
   yx -= view[2]/2;
   zx -= view[2]/2;

   xy -= view[3]/2;
   yy -= view[3]/2;
   zy -= view[3]/2;

   
   #define abs(x) ((x >= 0) ? x : -x)
   #define sign(x) ((x >= 0) ? 1 : -1)
   if(abs(xx) >= abs(yx) && abs(xx) >= abs(zx))
   {
      right = 0;
      sright = sign(xx);
   }
   else if(abs(zx) >= abs(yx) && abs(zx) >= abs(xx))
   {
      right = 2;
      sright = sign(zx);
   }
   else
   {
      right = 1;
      sright = sign(yx);
   }

   if(abs(yy) >= abs(xy) && abs(yy) >= abs(zy))
   {
      up = 1;
      sup = sign(yy);

      if(right == 0)
      {
         bottom = 2;
         sbottom = -sright;
      }
      else
      {
         bottom = 0;
         sbottom = sright;
      }
   }
   else if(abs(zy) >= abs(xy) && abs(zy) >= abs(yy))
   {
      up = 2;
      sup = sign(zy);

      if(right == 0)
      {
         bottom = 1;
         sbottom = -sright;
      }
      else
      {
         bottom = 0;
         sbottom = sright;
      }
   }
   else
   {
      up = 0;
      sup = sign(xy);

      if(right == 1)
      {
         bottom = 2;
         sbottom = sright;
      }
      else
      {
         bottom = 1;
         sbottom = -sright;
      }
   }


   //glGetIntegerv(GL_VIEWPORT,view);
  
   scale = view[2]*SCALE_FACTOR; 

//   result[0] = p[0]+(double)x/scale;
   result[right] = p[right]+(double)x/scale*sright;
   if(_buttons[MTKX_MIDDLE_BUTTON] == MTKX_DOWN)
   {
//      result[1] = p[1];
      result[up] = p[up];
//      result[2] = p[2]-(double)y/scale;
      result[bottom] = p[bottom]+(double)y/scale*sbottom;
   }
   else
   {
//      result[1] = p[1]+(double)y/scale;
      result[up] = p[up]+(double)y/scale*sup;
//      result[2] = p[2];
      result[bottom] = p[bottom];
   }
}

mtkxGlutWindow::mtkxGlutWindow(mtkCoreGL *core, mtkCameraId cid, 
                               char *nome) :  mtkxWindow(core,cid)
{
  _mtkxGlutWindowID = glutCreateWindow(nome);
  _next = _mtkxGlutWindowList;
  _mtkxGlutWindowList = this;

  _core->cameraInitialize(_cid);
  glutReshapeFunc(_mtkxReshape);
  glutDisplayFunc(_mtkxDisplay);
  glutMouseFunc(_mtkxMouse);
  glutMotionFunc(_mtkxMotion);
  glutPassiveMotionFunc(_mtkxPassiveMotion);

/*
  if(glutDeviceGet(GLUT_HAS_SPACEBALL))
     printf("Tem ");
  else
     printf("Nao tem ");
  glutSpaceballMotionFunc(_mtkxSpaceballMotion);
  printf("Spaceball.\n");
*/

  glutEntryFunc(_mtkxEntry);

  _warped = 0;
}

void mtkxGlutWindow::mtkxGlutInitWindow(int *argc, char **argv)
{
   glutInit(argc, argv);
   glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
   _buttons[MTKX_LEFT_BUTTON] = MTKX_UP;
   _buttons[MTKX_MIDDLE_BUTTON] = MTKX_UP;
   _buttons[MTKX_RIGHT_BUTTON] = MTKX_UP;

}

void mtkxGlutWindow::mtkxMainLoop()
{
   glutMainLoop();
}

void mtkxGlutWindow::mtkxSetCursorType(mtkxCursorType c)
{
   mtkxWindow::mtkxSetCursorType(c);
   glutSetWindow(_mtkxGlutWindowID);
   if(c == MTKX_CURSOR_2D)
   {
      glutSetCursor(GLUT_CURSOR_INHERIT);
      _core->deactivateCursor();
   }
   else if(c == MTKX_CURSOR_3D)
   {
      glutSetCursor(GLUT_CURSOR_NONE);
   //   _core->activateCursor();
      _entered = 1;
   }
   mtkxPostRedisplay();
}

void mtkxGlutWindow::mtkxPostRedisplay()
{
   glutPostWindowRedisplay(_mtkxGlutWindowID);
}

void mtkxGlutWindow::mtkxKeyboardFunc(mtkxKeyboardCallback keyboard)
{
   mtkxWindow::mtkxKeyboardFunc(keyboard);
   glutSetWindow(_mtkxGlutWindowID);
   glutKeyboardFunc(keyboard);
}

mtkxGlutWindow *mtkxGlutWindow::_getCurrentWindow()
{
   int ID = glutGetWindow();
   mtkxGlutWindow *p = _mtkxGlutWindowList;

   while(p && p->_mtkxGlutWindowID != ID)
      p = p->_next;

   return p;
}

void mtkxGlutWindow::_mtkxReshape(int w, int h)
{
   mtkxGlutWindow *p = _getCurrentWindow();
   if(p)
      p->mtkxReshape(w,h);
}

void mtkxGlutWindow::_mtkxDisplay()
{
   mtkxGlutWindow *p = _getCurrentWindow();
   if(p)
      p->mtkxDisplay();
}

void mtkxGlutWindow::_mtkxMouse(int button, int state, int x, int y)
{
   mtkxGlutWindow *p = _getCurrentWindow();
   if(p)
      p->mtkxMouse(button, state, x, y);
}

void mtkxGlutWindow::_mtkxMotion(int x, int y)
{
   mtkxGlutWindow *p = _getCurrentWindow();
   if(p)
      p->mtkxMotion(x, y);
}

void mtkxGlutWindow::_mtkxPassiveMotion(int x, int y)
{
   mtkxGlutWindow *p = _getCurrentWindow();
   if(p)
      p->mtkxPassiveMotion(x, y);
}

/*
void mtkxGlutWindow::_mtkxSpaceballMotion(int x, int y, int z)
{
   mtkxGlutWindow *p = _getCurrentWindow();
   if(p)
      p->mtkxSpaceballMotion(x, y, z);
}
*/

void mtkxGlutWindow::_mtkxEntry(int entrou)
{
   mtkxGlutWindow *p = _getCurrentWindow();
   if(p)
      p->mtkxEntry(entrou);
}

void mtkxGlutWindow::mtkxReshape(int w, int h)
{
   if(_reshapeCallback)
      (*_reshapeCallback)(w,h);
   else
      if(_core)
         _core->cameraResize(_cid, 0, 0, w, h);
}

void mtkxGlutWindow::mtkxDisplay() 
{
   if(_displayCallback)
      (*_displayCallback)();
   else
      if(_core)
         _core->cameraDraw(_cid);
   glutSwapBuffers();
}

void mtkxGlutWindow::mtkxMouse(int button, int state, int x, int y)
{
   gmPoint3 position;
   int botao, estado;

   estado = (state == GLUT_DOWN) ? MTKX_DOWN : MTKX_UP;

   switch(button)
   {
      case GLUT_LEFT_BUTTON:
         botao = MTKX_LEFT_BUTTON;
         break;

      case GLUT_MIDDLE_BUTTON:
         botao = MTKX_MIDDLE_BUTTON;
         break;

      case GLUT_RIGHT_BUTTON:
         botao = MTKX_RIGHT_BUTTON;
         break;

      default:
         botao = MTKX_LEFT_BUTTON;
         break;
   }

   _buttons[botao] = estado;

   if(_CursorType == MTKX_CURSOR_2D)
   {   
      if(_mouseCallback != NULL)
         (*_mouseCallback)(botao, estado, x, y);
      else
      {
         if(_core)
         {
            if(estado == MTKX_DOWN)
               _core->cameraPointer(_cid,x,y,MTK_PICK); 
            else
               _core->cameraPointer(_cid,x,y,MTK_RELEASE); 
            mtkxPostRedisplay();
         }
      }
   }
   else
   {
      if(_mouse3DCallback != NULL)
      {
	 position = _core->getCursorPosition();
	 (*_mouse3DCallback)(botao, estado, position[0],
                            position[1],
                            position[2]);
      }
   }   
}

void mtkxGlutWindow::mtkxMotion(int x, int y)
{
   if(_CursorType == MTKX_CURSOR_2D)
   {
      if(_motionCallback != NULL)
         (*_motionCallback)(x,y);
      else
         if(_core)
         {
            _core->cameraPointer(_cid,x,y,MTK_MOVE);
            mtkxPostRedisplay();
         }
   }
   else
      mtkxHandle3DMotion(x, y, _motion3DCallback);
}

void mtkxGlutWindow::mtkxPassiveMotion(int x, int y)
{
   if(_CursorType == MTKX_CURSOR_2D)
   {
      if(_passiveMotionCallback != NULL)
         (*_passiveMotionCallback)(x,y);
   }
   else
      mtkxHandle3DMotion(x, y, _passiveMotion3DCallback);
}

/*
void mtkxGlutWindow::mtkxSpaceballMotion(int x, int y, int z)
{
   gmPoint3 position, delta;
   GLdouble scale;
   GLint view[4];

   if(_CursorType == MTKX_CURSOR_3D)
   {
      position = _core->getCursorPosition();
      glGetIntegerv(GL_VIEWPORT,view);
  
      scale = view[2]*SCALE_FACTOR; 
      delta[0] = position[0] + (double)x/scale;
      delta[1] = position[1] + (double)y/scale;
      delta[2] = position[2] + (double)z/scale;

      _core->cameraPointer3D(_cid,delta[0],delta[1],delta[2],MTK_MOVE);

      printf("Mexeu!\n");

      mtkxPostRedisplay();
   }
}
*/

void mtkxGlutWindow::mtkxHandle3DMotion(int x, int y, mtkxMotion3DCallback cb)
{
   GLdouble cx, cy, cz;
   GLdouble model[16], proj[16];
   GLint view[4];
   gmPoint3 position, delta;

/*   static int warped = 0; 
   static int _ax,_ay;
   static int lx, ly; */

   if(_core)
   {
      //printf("x:%d y:%d lx:%d ly:%d \n",x,y,lx,ly);
      if(_warped && x == _lx && y == _ly)
      {
         _ax = x;
         _ay = y;
         _warped = 0;
         return;
      }

      glGetDoublev(GL_MODELVIEW_MATRIX,model);
      glGetDoublev(GL_PROJECTION_MATRIX,proj);
      glGetIntegerv(GL_VIEWPORT,view);

      position = _core->getCursorPosition();

      // obtain 3D cursor new position
      if(_entered)
      {
         _entered = 0;
         _warped = 0;

         _core->activateCursor();

	 gluProject(position[0],position[1],position[2], model, 
                    proj, view, &cx,&cy,&cz);
             
	 gluUnProject(x, view[3]-y, cz, model, proj, view, &cx, &cy,
                      &cz);
      }
      else
      {
         if(_motionMapping != NULL)
 	    (*_motionMapping)(position, x-_ax, _ay-y, delta);
         else if(_MotionType == MTKX_MOTION_IMPLICIT)
            _mtkxMappingImplicit(position, x-_ax, _ay-y, delta, _implicit);
         else
         {
            delta[0] = delta[1] = delta[2] = 0;
         }

	 cx = delta[0];
	 cy = delta[1];
	 cz = delta[2];
      }

      if(cb != NULL)
         (*cb)(cx,cy,cz);
      else
      {
	// Move 3D pointer
        _core->cameraPointer3D(_cid,cx,cy,cz,MTK_MOVE);
      }

      // update cursor position
      gluProject(cx,cy,cz, model, proj, view, &cx,&cy,&cz);

      if(!_warped)
      {
	 _lx = (int)cx;
	 _ly = view[3]-(int)cy;
         
  	 glutWarpPointer(_lx,_ly);
        
         _warped = 1;
      }
      _ax = x; _ay = y;

      mtkxPostRedisplay();
   }
}

void mtkxGlutWindow::mtkxEntry(int entrou)
{
   if(entrou == GLUT_ENTERED)
   {
      _entered = 1;
      if(_core && _CursorType == MTKX_CURSOR_3D)
      {
//         _core->activateCursor();
         mtkxPostRedisplay();
      }
   }
   else
   {
      _entered = 0;
      if(_core && _CursorType == MTKX_CURSOR_3D)
      {
         _core->deactivateCursor();
         mtkxPostRedisplay();
      }
   }
}


