// -*- mode: C++; -*-
 
/* Manipulation Toolkit Extension (MTKX)
 * Copyright (C) 2000 Leonardo A. G. Mesquita
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef MTKX
#define MTKX

#include "gm.h"
#include "mtk.h"
#include <gtk/gtk.h>

class mtkxWindow {

public:
  typedef enum {MTKX_CURSOR_2D,MTKX_CURSOR_3D} mtkxCursorType;
  typedef enum {MTKX_MOTION_USER,MTKX_MOTION_PLANES,
		MTKX_MOTION_IMPLICIT} mtkxMotionType;
  static const int MTKX_UP = 0;
  static const int MTKX_DOWN = 1;
  static const int MTKX_LEFT_BUTTON = 0;  
  static const int MTKX_MIDDLE_BUTTON = 1;  
  static const int MTKX_RIGHT_BUTTON = 2;

protected:
  typedef void mtkxReshapeCallback (int, int);
  typedef void mtkxDisplayCallback ();
  typedef void mtkxKeyboardCallback (unsigned char,int,int);
  typedef void mtkxMouseCallback (int,int,int,int);
  typedef void mtkxMouse3DCallback (int,int,GLdouble,GLdouble,GLdouble);
  typedef void mtkxMotionCallback (int,int);
  typedef void mtkxMotion3DCallback (GLdouble,GLdouble,GLdouble);
  typedef void mtkxPassiveMotionCallback (int,int);
  typedef void mtkxPassiveMotion3DCallback (GLdouble,GLdouble,GLdouble);
  typedef void mtkxMotionMapping (gmPoint3,int,int,gmPoint3 &);
  typedef GLdouble mtkxImplicitFunction(GLdouble, GLdouble);

  mtkxCursorType _CursorType;
  mtkxMotionType _MotionType;
  mtkCoreGL *_core;
  mtkCameraId _cid;

  mtkxReshapeCallback *_reshapeCallback;
  mtkxDisplayCallback *_displayCallback;
  mtkxKeyboardCallback *_keyboardCallback;
  mtkxMouseCallback *_mouseCallback;
  mtkxMouse3DCallback *_mouse3DCallback;
  mtkxMotionCallback *_motionCallback;
  mtkxMotion3DCallback *_motion3DCallback;
  mtkxPassiveMotionCallback *_passiveMotionCallback;
  mtkxPassiveMotion3DCallback *_passiveMotion3DCallback;
  mtkxMotionMapping *_motionMapping;
  mtkxImplicitFunction *_implicit;

  static int _buttons[3];  


public:
  static void _mtkxMappingPlanes(gmPoint3 p, int x, int y, gmPoint3 &result);
  static void _mtkxMappingImplicit(gmPoint3 p,int x, int y, gmPoint3 &result, 
                                   mtkxImplicitFunction *);

  mtkxWindow(mtkCoreGL *,mtkCameraId);

  virtual void mtkxSetCursorType(mtkxCursorType);
  virtual void mtkxSetMotionType(mtkxMotionType);
  void mtkxSetImplicit(mtkxImplicitFunction);

  virtual void mtkxReshapeFunc(mtkxReshapeCallback);
  virtual void mtkxDisplayFunc(mtkxDisplayCallback);
  virtual void mtkxKeyboardFunc(mtkxKeyboardCallback);
  virtual void mtkxMouseFunc(mtkxMouseCallback);
  virtual void mtkxMouse3DFunc(mtkxMouse3DCallback);
  virtual void mtkxMotionFunc(mtkxMotionCallback);
  virtual void mtkxMotion3DFunc(mtkxMotion3DCallback);
  virtual void mtkxPassiveMotionFunc(mtkxPassiveMotionCallback);
  virtual void mtkxPassiveMotion3DFunc(mtkxPassiveMotion3DCallback);
  virtual void mtkxMappingFunc(mtkxMotionMapping);

  virtual void mtkxPostRedisplay() = 0;
};


class mtkxGlutWindow : public mtkxWindow
{
public:
   
  mtkxGlutWindow(mtkCoreGL *, mtkCameraId, char *);
   
  static void mtkxGlutInitWindow(int *, char **);
  static void mtkxMainLoop();

  void mtkxSetCursorType(mtkxCursorType);
  void mtkxPostRedisplay();
  void mtkxKeyboardFunc(mtkxKeyboardCallback);

private:
  int _mtkxGlutWindowID;

  int _entered;
  int _warped;
  int _ax, _ay, _lx, _ly;

  static mtkxGlutWindow *_mtkxGlutWindowList;
  mtkxGlutWindow *_next;

  static mtkxGlutWindow *_getCurrentWindow();
  static void _mtkxReshape (int, int);
  static void _mtkxDisplay ();
  static void _mtkxMouse (int button, int state, int x, int y) ;
  static void _mtkxMotion (int x, int y);
  static void _mtkxPassiveMotion (int x, int y);
  //static void _mtkxSpaceballMotion (int x, int y, int z);
  static void _mtkxEntry (int entrou);

  void mtkxReshape (int, int);
  void mtkxDisplay ();
  void mtkxMouse (int button, int state, int x, int y) ;
  void mtkxMotion (int x, int y);
  void mtkxPassiveMotion (int x, int y);
  //void mtkxSpaceballMotion (int x, int y, int z);
  void mtkxHandle3DMotion (int x, int y, mtkxMotion3DCallback);
  void mtkxEntry (int entrou);
};

class mtkxGtkWindow : public mtkxWindow
{
public:
   
  mtkxGtkWindow(mtkCoreGL *, mtkCameraId, char *);
   
  static void mtkxGtkInitWindow(int *, char ***);
  static void mtkxMainLoop();

  void mtkxSetCursorType(mtkxCursorType);
  void mtkxPostRedisplay();

private:
  GtkWidget *_window, *_glarea;
  static GdkCursor *_vazio;

  int _entered;
  int _warped;
  int _initialized;
  int _ax, _ay, _lx, _ly;

  static mtkxGtkWindow *_mtkxGtkWindowList;
  mtkxGtkWindow *_next;

  static mtkxGtkWindow *_getCurrentWindow(GtkWidget *);
  static int _mtkxReshape (GtkWidget *, GdkEventConfigure *);
  static int _mtkxDisplay (GtkWidget *, GdkEventExpose *);
  static int _mtkxMouse (GtkWidget *, GdkEventButton *);
  static int _mtkxMotion (GtkWidget *, GdkEventMotion *);
  static int _mtkxEntry (GtkWidget *, GdkEventCrossing *);
  static int _mtkxKeyboard (GtkWidget *, GdkEventKey *);

  void mtkxReshape (int, int);
  void mtkxDisplay ();
  void mtkxMouse (int button, int state, int x, int y) ;
  void mtkxMotion (int x, int y);
  void mtkxPassiveMotion (int x, int y);
  void mtkxHandle3DMotion (int x, int y, mtkxMotion3DCallback);
  void mtkxEntry (int entrou);
};

#endif

