// -*- mode: C++; -*-
 
/* Manipulation Toolkit Extension (MTKX)
 * Copyright (C) 2000 Leonardo A. G. Mesquita
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <X11/Xlib.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkx.h>
#include <gdk/gdkkeysyms.h>
#include <gtkgl/gtkglarea.h>

#include <stdio.h>
#include "mtkx.h"


#define SCALE_FACTOR 0.24

mtkxGtkWindow *mtkxGtkWindow::_mtkxGtkWindowList;
GdkCursor *mtkxGtkWindow::_vazio;

mtkxGtkWindow::mtkxGtkWindow(mtkCoreGL *core, mtkCameraId cid, 
                               char *nome) :  mtkxWindow(core,cid)
{
  int attrlist[] = {
     GDK_GL_RGBA,
     GDK_GL_DOUBLEBUFFER,
     GDK_GL_DEPTH_SIZE, 1,
     GDK_GL_NONE
  };

  _next = _mtkxGtkWindowList;
  _mtkxGtkWindowList = this;
  _warped = 0;
  _initialized = 0;

  _window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(_window), nome);
  gtk_container_border_width(GTK_CONTAINER(_window), 0);
  gtk_signal_connect(GTK_OBJECT(_window), "delete_event",
                     GTK_SIGNAL_FUNC(gtk_main_quit), NULL);
  gtk_quit_add_destroy(1, GTK_OBJECT(_window));
  _glarea = GTK_WIDGET(gtk_gl_area_new(attrlist));
  gtk_widget_set_events(GTK_WIDGET(_glarea),
                        GDK_EXPOSURE_MASK |
                        GDK_BUTTON_PRESS_MASK |
                        GDK_POINTER_MOTION_MASK |
                        GDK_BUTTON_RELEASE_MASK |
                        GDK_ENTER_NOTIFY_MASK |
                        GDK_LEAVE_NOTIFY_MASK |
                        GDK_KEY_PRESS_MASK |
                        GDK_KEY_RELEASE_MASK);

  gtk_signal_connect(GTK_OBJECT(_glarea), "key_press_event",
                     GTK_SIGNAL_FUNC(_mtkxKeyboard), NULL);
  gtk_signal_connect(GTK_OBJECT(_glarea), "configure_event",
                     GTK_SIGNAL_FUNC(_mtkxReshape), NULL);
  gtk_signal_connect(GTK_OBJECT(_glarea), "expose_event",
                     GTK_SIGNAL_FUNC(_mtkxDisplay), NULL);
  gtk_signal_connect(GTK_OBJECT(_glarea), "button_press_event",
                     GTK_SIGNAL_FUNC(_mtkxMouse), NULL);
  gtk_signal_connect(GTK_OBJECT(_glarea), "button_release_event",
                     GTK_SIGNAL_FUNC(_mtkxMouse), NULL);
  gtk_signal_connect(GTK_OBJECT(_glarea), "motion_notify_event",
                     GTK_SIGNAL_FUNC(_mtkxMotion), NULL);
  gtk_signal_connect(GTK_OBJECT(_glarea), "enter_notify_event",
                     GTK_SIGNAL_FUNC(_mtkxEntry), NULL);
  gtk_signal_connect(GTK_OBJECT(_glarea), "leave_notify_event",
                     GTK_SIGNAL_FUNC(_mtkxEntry), NULL);


//  gtk_widget_set_usize(GTK_WIDGET(_glarea), 300, 300);
  gtk_gl_area_size(GTK_GL_AREA(_glarea), 300, 300);

  gtk_container_add(GTK_CONTAINER(_window), GTK_WIDGET(_glarea));
  gtk_widget_show(GTK_WIDGET(_glarea));
  gtk_widget_show(GTK_WIDGET(_window));

  GTK_WIDGET_SET_FLAGS(_glarea, GTK_CAN_FOCUS);
  gtk_widget_grab_focus(GTK_WIDGET(_glarea));
}

void mtkxGtkWindow::mtkxGtkInitWindow(int *argc, char ***argv)
{
   gtk_init(argc, argv);

   GdkPixmap *pix;
   GdkColor cor;
   char *data = new char[1];
   data[0] = 0;
   pix = gdk_bitmap_create_from_data(NULL, data, 1, 1);
   _vazio = gdk_cursor_new_from_pixmap(pix, pix, &cor, &cor, 0, 0);
   gdk_pixmap_unref(pix);
   delete data;

   _buttons[MTKX_LEFT_BUTTON] = MTKX_UP;
   _buttons[MTKX_MIDDLE_BUTTON] = MTKX_UP;
   _buttons[MTKX_RIGHT_BUTTON] = MTKX_UP;
}

void mtkxGtkWindow::mtkxMainLoop()
{
   gtk_main();
}

void mtkxGtkWindow::mtkxSetCursorType(mtkxCursorType c)
{
   mtkxWindow::mtkxSetCursorType(c);
   if(c == MTKX_CURSOR_2D)
   {
      gdk_window_set_cursor(_glarea->window,
           NULL);
      _core->deactivateCursor();
   }
   else if(c == MTKX_CURSOR_3D)
   {
      gdk_window_set_cursor(_glarea->window,
           _vazio);
      _core->activateCursor();
   }
}

void mtkxGtkWindow::mtkxPostRedisplay()
{
   gtk_widget_draw(GTK_WIDGET(_glarea), NULL);
}

mtkxGtkWindow *mtkxGtkWindow::_getCurrentWindow(GtkWidget *w)
{
   mtkxGtkWindow *p = _mtkxGtkWindowList;

   while(p && p->_glarea != w)
      p = p->_next;

   return p;
}


int mtkxGtkWindow::_mtkxKeyboard(GtkWidget *w, GdkEventKey *e)
{
   mtkxGtkWindow *p = _getCurrentWindow(w);
   if(p && p->_keyboardCallback)
   {
      (* (p->_keyboardCallback))(e->keyval, 0, 0);
   }
   return TRUE;
}

int mtkxGtkWindow::_mtkxReshape(GtkWidget *w, GdkEventConfigure *e)
{
   e = e;
   mtkxGtkWindow *p = _getCurrentWindow(w);
   if(p && gtk_gl_area_make_current(GTK_GL_AREA(w)))
   {
      p->mtkxReshape(w->allocation.width,w->allocation.height);
   }
   return TRUE;
}

int mtkxGtkWindow::_mtkxDisplay(GtkWidget *w, GdkEventExpose *e)
{
   e = e;
   mtkxGtkWindow *p = _getCurrentWindow(w);
   if(p && gtk_gl_area_make_current(GTK_GL_AREA(w)))
   {
      p->mtkxDisplay();
   }
   return TRUE;
}


int mtkxGtkWindow::_mtkxMouse(GtkWidget *w, GdkEventButton *e)
{
   mtkxGtkWindow *p = _getCurrentWindow(w);
   if(p)
      p->mtkxMouse(e->button, e->type, e->x, e->y);

   return TRUE;
}

int mtkxGtkWindow::_mtkxMotion(GtkWidget *w, GdkEventMotion *e)
{
   mtkxGtkWindow *p = _getCurrentWindow(w);
   if(p)
      if(e->state & (GDK_BUTTON1_MASK | GDK_BUTTON2_MASK |
                     GDK_BUTTON3_MASK) == 0)
         p->mtkxPassiveMotion(e->x, e->y);
      else
         p->mtkxMotion(e->x, e->y);
   return TRUE;
}


int mtkxGtkWindow::_mtkxEntry(GtkWidget *w, GdkEventCrossing *e)
{
   mtkxGtkWindow *p = _getCurrentWindow(w);
   if(p)
      p->mtkxEntry(e->type);
   return TRUE;
}

void mtkxGtkWindow::mtkxReshape(int w, int h)
{
   if(_reshapeCallback)
      (*_reshapeCallback)(w,h);
   else
      if(_core)
         _core->cameraResize(_cid, 0, 0, w, h);
}

void mtkxGtkWindow::mtkxDisplay() 
{
   if(_initialized == 0)
   {
      _core->cameraInitialize(_cid);
      _initialized = 1;
   }

   if(_displayCallback)
      (*_displayCallback)();
   else
      if(_core)
      {
         _core->cameraDraw(_cid);
         gtk_gl_area_swapbuffers(GTK_GL_AREA(_glarea));
      }
}

void mtkxGtkWindow::mtkxMouse(int button, int state, int x, int y)
{
   gmPoint3 position;
   int botao, estado;

   estado = (state == GDK_BUTTON_PRESS) ? MTKX_DOWN : MTKX_UP;

   switch(button)
   {
      case 1:
         botao = MTKX_LEFT_BUTTON;
         break;

      case 2:
         botao = MTKX_MIDDLE_BUTTON;
         break;

      case 3:
         botao = MTKX_RIGHT_BUTTON;
         break;

      default:
         botao = MTKX_LEFT_BUTTON;
         break;
   }

   _buttons[botao] = estado;

   if(_CursorType == MTKX_CURSOR_2D)
   {   
      if(_mouseCallback != NULL)
         (*_mouseCallback)(botao, estado, x, y);
      else
      {
         if(_core)
         {
            if(estado == MTKX_DOWN)
               _core->cameraPointer(_cid,x,y,MTK_PICK); 
            else
               _core->cameraPointer(_cid,x,y,MTK_RELEASE); 
            mtkxPostRedisplay();
         }
      }
   }
   else
   {
      if(_mouse3DCallback != NULL)
      {
	 position = _core->getCursorPosition();
	 (*_mouse3DCallback)(botao, estado, position[0],
                            position[1],
                            position[2]);
      }
   }   
}


void mtkxGtkWindow::mtkxMotion(int x, int y)
{
   if(_CursorType == MTKX_CURSOR_2D)
   {
      if(_motionCallback != NULL)
         (*_motionCallback)(x,y);
      else
         if(_core)
         {
            _core->cameraPointer(_cid,x,y,MTK_MOVE);
            mtkxPostRedisplay();
         }
   }
   else
      mtkxHandle3DMotion(x, y, _motion3DCallback);
}

void mtkxGtkWindow::mtkxPassiveMotion(int x, int y)
{
   if(_CursorType == MTKX_CURSOR_2D)
   {
      if(_passiveMotionCallback != NULL)
         (*_passiveMotionCallback)(x,y);
   }
   else
      mtkxHandle3DMotion(x, y, _passiveMotion3DCallback);
}

void mtkxGtkWindow::mtkxHandle3DMotion(int x, int y, mtkxMotion3DCallback cb)
{
   GLdouble cx, cy, cz;
   GLdouble model[16], proj[16];
   GLint view[4];
   gmPoint3 position, delta;

   if(_core)
   {
      //printf("x:%d y:%d lx:%d ly:%d \n",x,y,lx,ly);
      if(_warped && x == _lx && y == _ly)
      {
         _ax = x;
         _ay = y;
         _warped = 0;
         return;
      }

      glGetDoublev(GL_MODELVIEW_MATRIX,model);
      glGetDoublev(GL_PROJECTION_MATRIX,proj);
      glGetIntegerv(GL_VIEWPORT,view);

      position = _core->getCursorPosition();

      // obtain 3D cursor new position
      if(_entered)
      {
         _entered = 0;
         _warped = 0;
	 gluProject(position[0],position[1],position[2], model, 
                    proj, view, &cx,&cy,&cz);
             
	 gluUnProject(x, view[3]-y, cz, model, proj, view, &cx, &cy,
                      &cz);
      }
      else
      {
         if(_motionMapping != NULL)
 	    (*_motionMapping)(position, x-_ax, _ay-y, delta);
         else if(_MotionType == MTKX_MOTION_IMPLICIT)
            _mtkxMappingImplicit(position, x-_ax, _ay-y, delta, _implicit);
         else
         {
            delta[0] = delta[1] = delta[2] = 0;
         }

	 cx = delta[0];
	 cy = delta[1];
	 cz = delta[2];
      }

      if(cb != NULL)
         (*cb)(cx,cy,cz);
      else
      {
	// Move 3D pointer
        _core->cameraPointer3D(_cid,cx,cy,cz,MTK_MOVE);
      }

      // update cursor position
      gluProject(cx,cy,cz, model, proj, view, &cx,&cy,&cz);

      if(!_warped)
      {
	 _lx = (int)cx;
	 _ly = view[3]-(int)cy;
         
  	 XWarpPointer(GDK_WINDOW_XDISPLAY(_glarea->window), None,
  	              GDK_WINDOW_XWINDOW(_glarea->window), 0, 0, 0, 0,
                      _lx,_ly);
        
         _warped = 1;
      }
      _ax = x; _ay = y;

      mtkxPostRedisplay();
   }
}

void mtkxGtkWindow::mtkxEntry(int entrou)
{
   if(entrou == GDK_ENTER_NOTIFY)
   {
      _entered = 1;
      if(_core && _CursorType == MTKX_CURSOR_3D)
      {
         _core->activateCursor();
         mtkxPostRedisplay();
      }
   }
   else
   {
      _entered = 0;
      if(_core && _CursorType == MTKX_CURSOR_3D)
      {
         _core->deactivateCursor();
         mtkxPostRedisplay();
      }
   }
}


